# Note: Tests for the TitlesAlignments class are in blast/test_titles.py
#       and diamond/test_titles.py because that class needs a concrete
#       (iterable) dark.alignments.ReadsAlignments class passed to its
#       __init__.  The tests below test the simpler dark.titles classes,
#       TitleAlignment and TitleAlignments.

from collections import Counter
from unittest import TestCase

from dark.titles import TitleAlignment, TitleAlignments
from dark.reads import Read
from dark.hsp import HSP, LSP


class TestTitleAlignment(TestCase):
    """
    Test the TitleAlignment class.
    """

    def testExpectedAttributes(self):
        """
        An instance of TitleAlignment must have the expected attributes.
        """
        read = Read("id", "AAA")
        titleAlignment = TitleAlignment(read, [])
        self.assertEqual(read, titleAlignment.read)
        self.assertEqual([], titleAlignment.hsps)

    def testToDict(self):
        """
        The toDict method must return the expected result.
        """
        read = Read("the-id", "AAA")
        hsp1 = HSP(
            0,
            readStart=1,
            readEnd=2,
            readStartInSubject=3,
            readEndInSubject=4,
            subjectStart=5,
            subjectEnd=6,
            readMatchedSequence="aaa",
            subjectMatchedSequence="ccc",
            readFrame=7,
            subjectFrame=8,
            identicalCount=9,
            percentIdentical=31.3,
            positiveCount=10,
            percentPositive=4.5,
        )
        hsp2 = HSP(
            10,
            readStart=11,
            readEnd=12,
            readStartInSubject=13,
            readEndInSubject=14,
            subjectStart=15,
            subjectEnd=16,
            readMatchedSequence="ggg",
            subjectMatchedSequence="ttt",
            readFrame=17,
            subjectFrame=18,
            identicalCount=19,
            percentIdentical=32.3,
            positiveCount=20,
            percentPositive=4.6,
        )
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])

        self.assertEqual(
            {
                "hsps": [
                    {
                        "score": 0,
                        "readStart": 1,
                        "readEnd": 2,
                        "readStartInSubject": 3,
                        "readEndInSubject": 4,
                        "subjectStart": 5,
                        "subjectEnd": 6,
                        "readFrame": 7,
                        "subjectFrame": 8,
                        "identicalCount": 9,
                        "percentIdentical": 31.3,
                        "positiveCount": 10,
                        "percentPositive": 4.5,
                        "readMatchedSequence": "aaa",
                        "subjectMatchedSequence": "ccc",
                    },
                    {
                        "score": 10,
                        "readStart": 11,
                        "readEnd": 12,
                        "readStartInSubject": 13,
                        "readEndInSubject": 14,
                        "subjectStart": 15,
                        "subjectEnd": 16,
                        "readFrame": 17,
                        "subjectFrame": 18,
                        "identicalCount": 19,
                        "percentIdentical": 32.3,
                        "positiveCount": 20,
                        "percentPositive": 4.6,
                        "readMatchedSequence": "ggg",
                        "subjectMatchedSequence": "ttt",
                    },
                ],
                "read": {
                    "id": "the-id",
                    "quality": None,
                    "sequence": "AAA",
                },
            },
            titleAlignment.toDict(),
        )


class TestTitleAlignments(TestCase):
    """
    Test the TitleAlignments class.
    """

    def testExpectedAttributes(self):
        """
        An instance of TitleAlignments must have the expected attributes.
        """
        titleAlignments = TitleAlignments("subject title", 55)
        self.assertEqual("subject title", titleAlignments.subjectTitle)
        self.assertEqual(55, titleAlignments.subjectLength)
        self.assertEqual([], titleAlignments)

    def testAddAlignment(self):
        """
        It must be possible to add an alignment to an instance of
        TitleAlignments.
        """
        titleAlignments = TitleAlignments("subject title", 55)
        read = Read("id", "AAA")
        titleAlignment = TitleAlignment(read, [])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(read, titleAlignments[0].read)
        self.assertEqual([], titleAlignments[0].hsps)

    def testHSPs(self):
        """
        The hsps function must produce a list of all HSPs.
        """
        hsp1 = HSP(7)
        hsp2 = HSP(14)
        hsp3 = HSP(21)
        titleAlignments = TitleAlignments("subject title", 55)
        read = Read("id1", "AAA")
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read("id2", "AAA")
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(
            [7, 14, 21], [hsp.score.score for hsp in titleAlignments.hsps()]
        )

    def testReadsEmpty(self):
        """
        The reads function must return an empty Reads instance if there are no
        reads for the title.
        """
        titleAlignments = TitleAlignments("subject title", 55)
        self.assertEqual(0, len(list(titleAlignments.reads())))

    def testReads(self):
        """
        The reads function must return a Reads instance with the reads for
        the title.
        """
        hsp1 = HSP(7)
        hsp2 = HSP(14)
        hsp3 = HSP(21)
        titleAlignments = TitleAlignments("subject title", 55)
        read1 = Read("id1", "AAA")
        titleAlignment = TitleAlignment(read1, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read2 = Read("id2", "AAA")
        titleAlignment = TitleAlignment(read2, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual([read1, read2], list(titleAlignments.reads()))

    def testReadCountZero(self):
        """
        The readCount function must return zero if no reads matched a title.
        """
        titleAlignments = TitleAlignments("subject title", 55)
        self.assertEqual(0, titleAlignments.readCount())

    def testReadCount(self):
        """
        The readCount function must indicate how many reads matched a title.
        """
        hsp1 = HSP(7)
        hsp2 = HSP(14)
        hsp3 = HSP(21)
        titleAlignments = TitleAlignments("subject title", 55)
        read = Read("id1", "AAA")
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read("id2", "AAA")
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(2, titleAlignments.readCount())

    def testHspCountZero(self):
        """
        The hspCount function must return zero if no reads matched a title.
        """
        titleAlignments = TitleAlignments("subject title", 55)
        self.assertEqual(0, titleAlignments.hspCount())

    def testHspCount(self):
        """
        The hspCount function must indicate how many HSPs were found in
        total for all the alignments to a title.
        """
        hsp1 = HSP(7)
        hsp2 = HSP(14)
        hsp3 = HSP(21)
        titleAlignments = TitleAlignments("subject title", 55)
        read = Read("id1", "AAA")
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read("id2", "AAA")
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(3, titleAlignments.hspCount())

    def testMedianScoreWithNoAlignments(self):
        """
        The medianScore function must raise IndexError (due to no inputs)
        if there are no alignments matching a title.
        """
        titleAlignments = TitleAlignments("subject title", 55)
        error = "^arg is an empty sequence$"
        self.assertRaisesRegex(ValueError, error, titleAlignments.medianScore)

    def testMedianScoreWithNoHsps(self):
        """
        The medianScore function must raise ValueError if there are no HSPs.
        """
        titleAlignments = TitleAlignments("subject title", 55)
        read = Read("id1", "AAA")
        titleAlignment = TitleAlignment(read, [])
        titleAlignments.addAlignment(titleAlignment)
        error = "^arg is an empty sequence$"
        self.assertRaisesRegex(ValueError, error, titleAlignments.medianScore)

    def testMedianScoreOfTwo(self):
        """
        The medianScore function must return the median score for the HSPs in
        all the alignments matching a title when given 2 scores.
        """
        hsp1 = HSP(7)
        hsp2 = HSP(15)
        titleAlignments = TitleAlignments("subject title", 55)
        read = Read("id1", "AAA")
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(11, titleAlignments.medianScore())

    def testMedianScoreOfThree(self):
        """
        The medianScore function must return the median score for the HSPs in
        all the alignments matching a title when given 3 scores.
        """
        hsp1 = HSP(7)
        hsp2 = HSP(15)
        hsp3 = HSP(21)
        titleAlignments = TitleAlignments("subject title", 55)
        read = Read("id1", "AAA")
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read("id2", "AAA")
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(15, titleAlignments.medianScore())

    def testBestHspWithNoHsps(self):
        """
        The bestHsp function must raise ValueError if there are no HSPs.
        """
        titleAlignments = TitleAlignments("subject title", 55)
        read = Read("id1", "AAA")
        titleAlignment = TitleAlignment(read, [])
        titleAlignments.addAlignment(titleAlignment)
        read = Read("id2", "AAA")
        titleAlignment = TitleAlignment(read, [])
        titleAlignments.addAlignment(titleAlignment)
        error = "^max\\(\\) arg is an empty sequence$"
        self.assertRaisesRegex(ValueError, error, titleAlignments.bestHsp)

    def testBestHsp(self):
        """
        The bestHsp function must return the HSP with the best score for all
        the HSPs for all the alignments matching a title.
        """
        hsp1 = HSP(7)
        hsp2 = HSP(15)
        hsp3 = HSP(21)
        titleAlignments = TitleAlignments("subject title", 55)
        read = Read("id1", "AAA")
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read("id2", "AAA")
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(hsp3, titleAlignments.bestHsp())

    def testWorstHspWithNoHsps(self):
        """
        The worstHsp function must raise ValueError if there are no HSPs.
        """
        titleAlignments = TitleAlignments("subject title", 55)
        read = Read("id1", "AAA")
        titleAlignment = TitleAlignment(read, [])
        titleAlignments.addAlignment(titleAlignment)
        read = Read("id2", "AAA")
        titleAlignment = TitleAlignment(read, [])
        titleAlignments.addAlignment(titleAlignment)
        error = "^min\\(\\) arg is an empty sequence$"
        self.assertRaisesRegex(ValueError, error, titleAlignments.worstHsp)

    def testWorstHsp(self):
        """
        The worstHsp function must return the HSP with the worst score for all
        the HSPs for all the alignments matching a title.
        """
        hsp1 = HSP(7)
        hsp2 = HSP(15)
        hsp3 = HSP(21)
        titleAlignments = TitleAlignments("subject title", 55)
        read = Read("id1", "AAA")
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read("id2", "AAA")
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(hsp1, titleAlignments.worstHsp())

    def testBetterThanFalse(self):
        """
        The hasScoreBetterThan function must return False if there is no HSP
        with a score better than the passed value.
        """
        hsp1 = HSP(7)
        hsp2 = HSP(15)
        hsp3 = HSP(21)
        titleAlignments = TitleAlignments("subject title", 55)
        read = Read("id1", "AAA")
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read("id2", "AAA")
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertFalse(titleAlignments.hasScoreBetterThan(21))

    def testBetterThanTrue(self):
        """
        The hasScoreBetterThan function must return True if there is an HSP
        with a score better than the passed value.
        """
        hsp1 = HSP(7)
        hsp2 = HSP(15)
        hsp3 = HSP(21)
        titleAlignments = TitleAlignments("subject title", 55)
        read = Read("id1", "AAA")
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read("id2", "AAA")
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertTrue(titleAlignments.hasScoreBetterThan(20))

    def testCoverageNoReads(self):
        """
        The coverage method must return zero when a title alignments has no
        alignments (and therefore no coverage).
        """
        titleAlignments = TitleAlignments("subject title", 100)
        self.assertEqual(0.0, titleAlignments.coverage())

    def testFullCoverage(self):
        """
        The coverage method must return the correct value when the title is
        fully covered by its reads.
        """
        hsp1 = HSP(7, subjectStart=0, subjectEnd=50)
        hsp2 = HSP(8, subjectStart=50, subjectEnd=100)
        titleAlignments = TitleAlignments("subject title", 100)
        read = Read("id1", "AAA")
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(1.0, titleAlignments.coverage())

    def testPartialCoverage(self):
        """
        The coverage method must return the correct value when the title is
        partially covered by its reads.
        """
        hsp1 = HSP(7, subjectStart=10, subjectEnd=20)
        hsp2 = HSP(15, subjectStart=30, subjectEnd=40)
        hsp3 = HSP(21, subjectStart=50, subjectEnd=60)
        titleAlignments = TitleAlignments("subject title", 100)
        read = Read("id1", "AAA")
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read("id2", "AAA")
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(0.3, titleAlignments.coverage())

    def testFullCoverageCounts(self):
        """
        The coverageCounts method must return the correct result when the title
        is fully covered by its reads.
        """
        hsp1 = HSP(7, subjectStart=0, subjectEnd=5)
        hsp2 = HSP(8, subjectStart=5, subjectEnd=10)
        titleAlignments = TitleAlignments("subject title", 10)
        read = Read("id1", "AAA")
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        c = Counter([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        self.assertEqual(c, titleAlignments.coverageCounts())

    def testCoverageCounts(self):
        """
        The coverageCounts method must return the correct results when the
        title is partially covered by its reads.
        """
        hsp1 = HSP(7, subjectStart=1, subjectEnd=2)
        hsp2 = HSP(15, subjectStart=3, subjectEnd=4)
        hsp3 = HSP(21, subjectStart=5, subjectEnd=6)
        titleAlignments = TitleAlignments("subject title", 10)
        read = Read("id1", "AAA")
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read("id2", "AAA")
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        c = Counter([1, 3, 5])
        self.assertEqual(c, titleAlignments.coverageCounts())

    def testCoverageCountsOverlap(self):
        """
        The coverageCounts method must return the correct results when the
        title is partially covered by its reads that overlap.
        """
        hsp1 = HSP(7, subjectStart=1, subjectEnd=2)
        hsp2 = HSP(15, subjectStart=3, subjectEnd=6)
        hsp3 = HSP(21, subjectStart=5, subjectEnd=6)
        titleAlignments = TitleAlignments("subject title", 10)
        read = Read("id1", "AAA")
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read("id2", "AAA")
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        c = Counter([1, 3, 4, 5, 5])
        self.assertEqual(c, titleAlignments.coverageCounts())

    def testCoverageInfoNoReads(self):
        """
        When a title has no reads aligned to it, the coverageInfo method
        must return an empty result.
        """
        titleAlignments = TitleAlignments("subject title", 55)
        coverage = titleAlignments.coverageInfo()
        self.assertEqual({}, coverage)

    def testCoverageInfoOneReadWithOneHSP(self):
        """
        When a title has one read with one HSP aligned to it, the coverageInfo
        method must return just the indices and bases from that read.
        """
        titleAlignments = TitleAlignments("subject title", 55)
        hsp = HSP(15, subjectStart=3, subjectEnd=6, readMatchedSequence="CGT")
        read = Read("id1", "AAACGT")
        titleAlignment = TitleAlignment(read, [hsp])
        titleAlignments.addAlignment(titleAlignment)
        coverage = titleAlignments.coverageInfo()
        self.assertEqual(
            {
                3: [(15, "C")],
                4: [(15, "G")],
                5: [(15, "T")],
            },
            coverage,
        )

    def testCoverageInfoOneReadWithTwoHSPs(self):
        """
        When a title has one read with two HSPs aligned to it, the coverageInfo
        method must return the correct indices and bases from that read.
        """
        titleAlignments = TitleAlignments("subject title", 55)
        hsp1 = HSP(15, subjectStart=1, subjectEnd=4, readMatchedSequence="A-A")
        hsp2 = HSP(10, subjectStart=3, subjectEnd=6, readMatchedSequence="CGT")
        read = Read("id1", "AAACGT")
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        coverage = titleAlignments.coverageInfo()
        self.assertEqual(
            {
                1: [(15, "A")],
                2: [(15, "-")],
                3: [(15, "A"), (10, "C")],
                4: [(10, "G")],
                5: [(10, "T")],
            },
            coverage,
        )

    def testCoverageInfoTwoReadsWithThreeHSPs(self):
        """
        When a title has two reads (one with two HSPs, one with one) aligned
        to it, the coverageInfo method must return the correct indices and
        bases from the read.
        """
        titleAlignments = TitleAlignments("subject title", 55)

        # First read.
        hsp1 = HSP(15, subjectStart=1, subjectEnd=4, readMatchedSequence="A-A")
        hsp2 = HSP(10, subjectStart=3, subjectEnd=6, readMatchedSequence="CGT")
        read = Read("id1", "AAACGT")
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)

        # Second read.
        hsp1 = HSP(20, subjectStart=5, subjectEnd=10, readMatchedSequence="CGGTA")
        read = Read("id2", "AAACGTCGGTAAAA")
        titleAlignment = TitleAlignment(read, [hsp1])
        titleAlignments.addAlignment(titleAlignment)

        coverage = titleAlignments.coverageInfo()
        self.assertEqual(
            {
                1: [(15, "A")],
                2: [(15, "-")],
                3: [(15, "A"), (10, "C")],
                4: [(10, "G")],
                5: [(10, "T"), (20, "C")],
                6: [(20, "G")],
                7: [(20, "G")],
                8: [(20, "T")],
                9: [(20, "A")],
            },
            coverage,
        )

    def testResidueCountsNoReads(self):
        """
        When a title has no reads aligned to it, the residueCounts method
        must return an empty result.
        """
        titleAlignments = TitleAlignments("subject title", 55)
        counts = titleAlignments.residueCounts()
        self.assertEqual(0, len(counts))

    def testResidueCountsUnknownCaseConversion(self):
        """
        The residueCounts method must raise a ValueError when asked to do an
        unknown case conversion.
        """
        titleAlignments = TitleAlignments("subject title", 55)
        error = "convertCaseTo must be one of 'none', 'lower', or 'upper'"
        self.assertRaisesRegex(
            ValueError, error, titleAlignments.residueCounts, convertCaseTo="xxx"
        )

    def testResidueCountsOneReadOneHSP(self):
        """
        The residueCounts method must return the correct result when just one
        read with one HSP is aligned to a title.
        """
        read = Read("id", "ACGT")
        hsp = HSP(
            33,
            readStart=0,
            readEnd=4,
            readStartInSubject=0,
            readEndInSubject=4,
            subjectStart=0,
            subjectEnd=4,
            readMatchedSequence="ACGT",
            subjectMatchedSequence="ACGT",
        )
        titleAlignments = TitleAlignments("subject title", 55)
        titleAlignment = TitleAlignment(read, [hsp])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(
            {
                0: {"A": 1},
                1: {"C": 1},
                2: {"G": 1},
                3: {"T": 1},
            },
            titleAlignments.residueCounts(),
        )

    def testResidueCountsOneReadOneHSPPartialMatch(self):
        """
        The residueCounts method must return the correct result when just one
        read with one HSP is aligned to a title and only part of the read
        matched the subject (all the read bases are still counted and
        returned).
        """
        read = Read("id", "ACGT")
        hsp = HSP(
            33,
            readStart=0,
            readEnd=2,
            readStartInSubject=0,
            readEndInSubject=4,
            subjectStart=0,
            subjectEnd=4,
            readMatchedSequence="ACGT",
            subjectMatchedSequence="ACGT",
        )
        titleAlignments = TitleAlignments("subject title", 55)
        titleAlignment = TitleAlignment(read, [hsp])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(
            {
                0: {"A": 1},
                1: {"C": 1},
                2: {"G": 1},
                3: {"T": 1},
            },
            titleAlignments.residueCounts(),
        )

    def testResidueCountsOneReadTwoHSPsAtStartOfSubject(self):
        """
        The residueCounts method must return the correct result when just one
        read with two HSPs is aligned to a title and the leftmost HSP is
        aligned with the left edge of the subject.

        HSP1:       ACGT
        HSP2:        CGTT
        """
        read = Read("id", "ACGT")
        hsp1 = HSP(
            33,
            readStart=0,
            readEnd=4,
            readStartInSubject=0,
            readEndInSubject=4,
            subjectStart=0,
            subjectEnd=4,
            readMatchedSequence="ACGT",
            subjectMatchedSequence="ACGT",
        )
        hsp2 = HSP(
            33,
            readStart=0,
            readEnd=4,
            readStartInSubject=1,
            readEndInSubject=5,
            subjectStart=1,
            subjectEnd=5,
            readMatchedSequence="CGTT",
            subjectMatchedSequence="CGTT",
        )
        titleAlignments = TitleAlignments("subject title", 55)
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(
            {
                0: {"A": 1},
                1: {"C": 2},
                2: {"G": 2},
                3: {"T": 2},
                4: {"T": 1},
            },
            titleAlignments.residueCounts(),
        )

    def testResidueCountsOneReadTwoHSPsNotAtStartOfSubject(self):
        """
        The residueCounts method must return the correct result when just one
        read with two HSPs is aligned to a title and the leftmost HSP is not
        aligned with the left edge of the subject.

        HSP1:       ACGT
        HSP2:        CGTT
        """
        read = Read("id", "ACGT")
        hsp1 = HSP(
            33,
            readStart=0,
            readEnd=4,
            readStartInSubject=10,
            readEndInSubject=14,
            subjectStart=10,
            subjectEnd=14,
            readMatchedSequence="ACGT",
            subjectMatchedSequence="ACGT",
        )
        hsp2 = HSP(
            33,
            readStart=0,
            readEnd=4,
            readStartInSubject=11,
            readEndInSubject=15,
            subjectStart=11,
            subjectEnd=15,
            readMatchedSequence="CGTT",
            subjectMatchedSequence="CGTT",
        )
        titleAlignments = TitleAlignments("subject title", 55)
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(
            {
                10: {"A": 1},
                11: {"C": 2},
                12: {"G": 2},
                13: {"T": 2},
                14: {"T": 1},
            },
            titleAlignments.residueCounts(),
        )

    def testResidueCountsNoCaseConversion(self):
        """
        The residueCounts method must return the correct result when asked not
        to convert case.

        HSP1:       AcgT
        HSP2:        CGTT
        """
        read = Read("id", "ACGT")
        hsp1 = HSP(
            33,
            readStart=0,
            readEnd=4,
            readStartInSubject=10,
            readEndInSubject=14,
            subjectStart=10,
            subjectEnd=14,
            readMatchedSequence="AcgT",
            subjectMatchedSequence="ACGT",
        )
        hsp2 = HSP(
            33,
            readStart=0,
            readEnd=4,
            readStartInSubject=11,
            readEndInSubject=15,
            subjectStart=11,
            subjectEnd=15,
            readMatchedSequence="CGTT",
            subjectMatchedSequence="CGTT",
        )
        titleAlignments = TitleAlignments("subject title", 55)
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(
            {
                10: {"A": 1},
                11: {"C": 1, "c": 1},
                12: {"G": 1, "g": 1},
                13: {"T": 2},
                14: {"T": 1},
            },
            titleAlignments.residueCounts(convertCaseTo="none"),
        )

    def testResidueCountsCaseConvertLower(self):
        """
        The residueCounts method must return the correct result when asked to
        convert residues to lower case.

        HSP1:       AcgT
        HSP2:        CGTT
        """
        read = Read("id", "ACGT")
        hsp1 = HSP(
            33,
            readStart=0,
            readEnd=4,
            readStartInSubject=10,
            readEndInSubject=14,
            subjectStart=10,
            subjectEnd=14,
            readMatchedSequence="AcgT",
            subjectMatchedSequence="ACGT",
        )
        hsp2 = HSP(
            33,
            readStart=0,
            readEnd=4,
            readStartInSubject=11,
            readEndInSubject=15,
            subjectStart=11,
            subjectEnd=15,
            readMatchedSequence="CGTT",
            subjectMatchedSequence="CGTT",
        )
        titleAlignments = TitleAlignments("subject title", 55)
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(
            {
                10: {"a": 1},
                11: {"c": 2},
                12: {"g": 2},
                13: {"t": 2},
                14: {"t": 1},
            },
            titleAlignments.residueCounts(convertCaseTo="lower"),
        )

    def testResidueCountsCaseConvertUpper(self):
        """
        The residueCounts method must return the correct result when asked to
        convert residues to upper case.

        HSP1:       AcgT
        HSP2:        CGTT
        """
        read = Read("id", "ACGT")
        hsp1 = HSP(
            33,
            readStart=0,
            readEnd=4,
            readStartInSubject=10,
            readEndInSubject=14,
            subjectStart=10,
            subjectEnd=14,
            readMatchedSequence="AcgT",
            subjectMatchedSequence="ACGT",
        )
        hsp2 = HSP(
            33,
            readStart=0,
            readEnd=4,
            readStartInSubject=11,
            readEndInSubject=15,
            subjectStart=11,
            subjectEnd=15,
            readMatchedSequence="CGTT",
            subjectMatchedSequence="CGTT",
        )
        titleAlignments = TitleAlignments("subject title", 55)
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(
            {
                10: {"A": 1},
                11: {"C": 2},
                12: {"G": 2},
                13: {"T": 2},
                14: {"T": 1},
            },
            titleAlignments.residueCounts(convertCaseTo="upper"),
        )

    def testResidueCountsCaseConvertUpperIsDefault(self):
        """
        The residueCounts method must convert to uppercase by default.

        HSP1:       AcgT
        HSP2:        CGTT
        """
        read = Read("id", "ACGT")
        hsp1 = HSP(
            33,
            readStart=0,
            readEnd=4,
            readStartInSubject=10,
            readEndInSubject=14,
            subjectStart=10,
            subjectEnd=14,
            readMatchedSequence="AcgT",
            subjectMatchedSequence="ACGT",
        )
        hsp2 = HSP(
            33,
            readStart=0,
            readEnd=4,
            readStartInSubject=11,
            readEndInSubject=15,
            subjectStart=11,
            subjectEnd=15,
            readMatchedSequence="CGTT",
            subjectMatchedSequence="CGTT",
        )
        titleAlignments = TitleAlignments("subject title", 55)
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(
            {
                10: {"A": 1},
                11: {"C": 2},
                12: {"G": 2},
                13: {"T": 2},
                14: {"T": 1},
            },
            titleAlignments.residueCounts(),
        )

    def testResidueCountsTwoReadsTwoHSPsLeftOverhang(self):
        """
        The residueCounts method must return the correct result when two
        reads, each with one HSP are aligned to a title and the leftmost HSP
        is aligned before the left edge of the subject (i.e, will include
        negative subject offsets).

        Subject:      GTT
        HSP1:       ACGT
        HSP2:        CGTT
        """
        read1 = Read("id", "ACGT")
        hsp1 = HSP(
            33,
            readStart=0,
            readEnd=4,
            readStartInSubject=-2,
            readEndInSubject=2,
            subjectStart=0,
            subjectEnd=2,
            readMatchedSequence="GT",
            subjectMatchedSequence="GT",
        )
        read2 = Read("id", "CGTT")
        hsp2 = HSP(
            33,
            readStart=0,
            readEnd=4,
            readStartInSubject=-1,
            readEndInSubject=3,
            subjectStart=0,
            subjectEnd=3,
            readMatchedSequence="GTT",
            subjectMatchedSequence="GTT",
        )
        titleAlignments = TitleAlignments("subject title", 55)
        titleAlignment = TitleAlignment(read1, [hsp1])
        titleAlignments.addAlignment(titleAlignment)
        titleAlignment = TitleAlignment(read2, [hsp2])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(
            {
                -2: {"A": 1},
                -1: {"C": 2},
                0: {"G": 2},
                1: {"T": 2},
                2: {"T": 1},
            },
            titleAlignments.residueCounts(),
        )

    def testResidueCountsOneReadTwoHSPsNotOverlapping(self):
        """
        The residueCounts method must return the correct result when just one
        read with two HSPs is aligned to a title and the HSPs do not overlap
        one another.

        HSP1:    ACGT
        HSP2:              CGTT
        """
        read = Read("id", "ACGT")
        hsp1 = HSP(
            33,
            readStart=0,
            readEnd=4,
            readStartInSubject=0,
            readEndInSubject=4,
            subjectStart=0,
            subjectEnd=4,
            readMatchedSequence="ACGT",
            subjectMatchedSequence="ACGT",
        )
        hsp2 = HSP(
            33,
            readStart=0,
            readEnd=4,
            readStartInSubject=10,
            readEndInSubject=14,
            subjectStart=10,
            subjectEnd=14,
            readMatchedSequence="CGTT",
            subjectMatchedSequence="CGTT",
        )
        titleAlignments = TitleAlignments("subject title", 55)
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(
            {
                0: {"A": 1},
                1: {"C": 1},
                2: {"G": 1},
                3: {"T": 1},
                10: {"C": 1},
                11: {"G": 1},
                12: {"T": 1},
                13: {"T": 1},
            },
            titleAlignments.residueCounts(),
        )

    def testSummaryWhenEmpty(self):
        """
        If summary is called on an instance of TitleAlignments with no
        alignments a ValueError must be raised.
        """
        titleAlignments = TitleAlignments("subject title", 55)
        error = "^max\\(\\) arg is an empty sequence$"
        self.assertRaisesRegex(ValueError, error, titleAlignments.summary)

    def testSummary(self):
        """
        The summary method must return the correct result.
        """
        titleAlignments = TitleAlignments("subject title", 10)
        titleAlignments.addAlignment(
            TitleAlignment(
                Read("id1", "ACGT"),
                [
                    HSP(30, subjectStart=0, subjectEnd=2),
                ],
            )
        )
        titleAlignments.addAlignment(
            TitleAlignment(
                Read("id2", "ACGT"),
                [
                    HSP(55, subjectStart=2, subjectEnd=4),
                    HSP(40, subjectStart=8, subjectEnd=9),
                ],
            )
        )
        self.assertEqual(
            {
                "bestScore": 55,
                "coverage": 0.5,
                "hspCount": 3,
                "medianScore": 40,
                "readCount": 2,
                "subjectLength": 10,
                "subjectTitle": "subject title",
            },
            titleAlignments.summary(),
        )

    def testToDict(self):
        """
        The toDict method must return the expected result.
        """
        read = Read("the-id", "AAA")
        hsp1 = HSP(
            0,
            readStart=1,
            readEnd=2,
            readStartInSubject=3,
            readEndInSubject=4,
            subjectStart=5,
            subjectEnd=6,
            readMatchedSequence="aaa",
            subjectMatchedSequence="ccc",
            readFrame=7,
            subjectFrame=8,
            identicalCount=9,
            percentIdentical=17.9,
            positiveCount=10,
            percentPositive=3.9,
        )
        hsp2 = HSP(
            10,
            readStart=11,
            readEnd=12,
            readStartInSubject=13,
            readEndInSubject=14,
            subjectStart=15,
            subjectEnd=16,
            readMatchedSequence="ggg",
            subjectMatchedSequence="ttt",
            readFrame=17,
            subjectFrame=18,
            identicalCount=19,
            percentIdentical=27.9,
            positiveCount=20,
            percentPositive=3.8,
        )
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments = TitleAlignments("subject title", 10)
        titleAlignments.addAlignment(titleAlignment)

        self.assertEqual(
            {
                "subjectTitle": "subject title",
                "subjectLength": 10,
                "titleAlignments": [
                    {
                        "hsps": [
                            {
                                "score": 0,
                                "readStart": 1,
                                "readEnd": 2,
                                "readStartInSubject": 3,
                                "readEndInSubject": 4,
                                "subjectStart": 5,
                                "subjectEnd": 6,
                                "readFrame": 7,
                                "subjectFrame": 8,
                                "identicalCount": 9,
                                "percentIdentical": 17.9,
                                "positiveCount": 10,
                                "percentPositive": 3.9,
                                "readMatchedSequence": "aaa",
                                "subjectMatchedSequence": "ccc",
                            },
                            {
                                "score": 10,
                                "readStart": 11,
                                "readEnd": 12,
                                "readStartInSubject": 13,
                                "readEndInSubject": 14,
                                "subjectStart": 15,
                                "subjectEnd": 16,
                                "readFrame": 17,
                                "subjectFrame": 18,
                                "identicalCount": 19,
                                "percentIdentical": 27.9,
                                "positiveCount": 20,
                                "percentPositive": 3.8,
                                "readMatchedSequence": "ggg",
                                "subjectMatchedSequence": "ttt",
                            },
                        ],
                        "read": {
                            "id": "the-id",
                            "quality": None,
                            "sequence": "AAA",
                        },
                    },
                ],
            },
            titleAlignments.toDict(),
        )


class TestTitleAlignmentsLSP(TestCase):
    """
    Test the TitleAlignments class using LSPs. The only tests here are ones
    that depend on lower scores being better.
    """

    def testBestHsp(self):
        """
        The bestHsp function must return the HSP with the best score for the
        HSPs all the alignments matching a title.
        """
        hsp1 = LSP(7)
        hsp2 = LSP(15)
        hsp3 = LSP(21)
        titleAlignments = TitleAlignments("subject title", 55)
        read = Read("id1", "AAA")
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read("id2", "AAA")
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(hsp1, titleAlignments.bestHsp())

    def testWorstHsp(self):
        """
        The worstHsp function must return the HSP with the worst score for all
        the HSPs for all the alignments matching a title.
        """
        hsp1 = LSP(7)
        hsp2 = LSP(15)
        hsp3 = LSP(21)
        titleAlignments = TitleAlignments("subject title", 55)
        read = Read("id1", "AAA")
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read("id2", "AAA")
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(hsp3, titleAlignments.worstHsp())

    def testBetterThanFalse(self):
        """
        The hasScoreBetterThan function must return False if there is no HSP
        with a score better than the passed value.
        """
        hsp1 = LSP(7)
        hsp2 = LSP(15)
        hsp3 = LSP(21)
        titleAlignments = TitleAlignments("subject title", 55)
        read = Read("id1", "AAA")
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read("id2", "AAA")
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertFalse(titleAlignments.hasScoreBetterThan(7))

    def testBetterThanTrue(self):
        """
        The hasScoreBetterThan function must return True if there is an HSP
        with a score better than the passed value.
        """
        hsp1 = LSP(7)
        hsp2 = LSP(15)
        hsp3 = LSP(21)
        titleAlignments = TitleAlignments("subject title", 55)
        read = Read("id1", "AAA")
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read("id2", "AAA")
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertTrue(titleAlignments.hasScoreBetterThan(9))

    def testReadIdsEmpty(self):
        """
        The readIds function must return the empty set if no reads matched a
        title.
        """
        titleAlignments = TitleAlignments("subject title", 55)
        self.assertEqual(0, len(titleAlignments.readIds()))

    def testReadIds(self):
        """
        The readIds function must return the set of read ids for the alignments
        matching a title.
        """
        hsp1 = LSP(7)
        hsp2 = LSP(15)
        hsp3 = LSP(21)
        titleAlignments = TitleAlignments("subject title", 55)
        read = Read("id1", "AAA")
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read("id2", "AAA")
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(set(["id1", "id2"]), titleAlignments.readIds())



from collections import defaultdict
from operator import itemgetter
from itertools import zip_longest

from dark.utils import countPrint

# A list of the ambiguous values is given at
# https://en.wikipedia.org/wiki/Nucleic_acid_notation
AMBIGUOUS: dict[str, set[str]] = {
    "A": {"A"},
    "C": {"C"},
    "G": {"G"},
    "T": {"T"},
    "M": {"A", "C"},
    "R": {"A", "G"},
    "W": {"A", "T"},
    "S": {"G", "C"},
    "K": {"G", "T"},
    "Y": {"C", "T"},
    "V": {"A", "C", "G"},
    "H": {"A", "C", "T"},
    "D": {"A", "G", "T"},
    "B": {"C", "G", "T"},
    "N": {"A", "C", "G", "T"},
}

# Make a reverse version of AMBIGUOUS.
BASES_TO_AMBIGUOUS = dict(
    ("".join(sorted(bases)), symbol) for symbol, bases in AMBIGUOUS.items()
)


def matchToString(
    dnaMatch,
    read1,
    read2,
    matchAmbiguous=True,
    indent="",
    offsets=None,
    includeGapLocations=True,
    includeNoCoverageLocations=True,
    includeAmbiguousMatches=False,
    includeNonGapMismatches=False,
):
    """
    Format a DNA match as a string.

    @param dnaMatch: A C{dict} returned by C{compareDNAReads}.
    @param read1: A C{Read} instance or an instance of one of its subclasses.
    @param read2: A C{Read} instance or an instance of one of its subclasses.
    @param matchAmbiguous: If C{True}, ambiguous nucleotides that are
        possibly correct were counted as actually being correct. Otherwise,
        the match was done strictly, insisting that only non-ambiguous
        nucleotides could contribute to the matching nucleotide count.
    @param indent: A C{str} to indent all returned lines with.
    @param offsets: If not C{None}, a C{set} of offsets of interest that were
        only considered when making C{match}.
    @param includeGapLocations: If C{True} indicate the (1-based) locations of
        gaps.
    @param includeNoCoverageLocations: If C{True} indicate the (1-based)
        locations of no coverage.
    @param includeAmbiguousMatches: If C{True} indicate the (1-based)
        locations of ambiguous matches.
    @param includeNonGapMismatches: If C{True} indicate the (1-based) locations
        of non-gap mismatches.
    @return: A C{str} describing the match.
    """
    match = dnaMatch["match"]
    identicalMatchCount = match["identicalMatchCount"]
    ambiguousMatchCount = match["ambiguousMatchCount"]
    gapMismatchCount = match["gapMismatchCount"]
    gapGapMismatchCount = match["gapGapMismatchCount"]
    nonGapMismatchCount = match["nonGapMismatchCount"]
    noCoverageCount = match["noCoverageCount"]
    noCoverageNoCoverageCount = match["noCoverageNoCoverageCount"]

    if offsets:
        len1 = len2 = len(offsets)
    else:
        len1, len2 = len(read1), len(read2)

    result = []
    append = result.append

    append(countPrint("%sExact matches" % indent, identicalMatchCount, len1, len2))
    append(countPrint("%sAmbiguous matches" % indent, ambiguousMatchCount, len1, len2))

    if noCoverageCount:
        append(
            countPrint(
                "%sExact matches (ignoring no coverage sites)" % indent,
                identicalMatchCount,
                len1 - noCoverageCount,
                len2 - noCoverageCount,
            )
        )
        append(
            countPrint(
                "%sAmbiguous matches (ignoring no coverage sites)" % indent,
                ambiguousMatchCount,
                len1 - noCoverageCount,
                len2 - noCoverageCount,
            )
        )

    if ambiguousMatchCount and identicalMatchCount:
        anyMatchCount = identicalMatchCount + ambiguousMatchCount
        append(
            countPrint(
                "%sExact or ambiguous matches" % indent, anyMatchCount, len1, len2
            )
        )
        if noCoverageCount:
            append(
                countPrint(
                    "%sExact or ambiguous matches (ignoring no coverage sites)"
                    % indent,
                    anyMatchCount,
                    len1 - noCoverageCount,
                    len2 - noCoverageCount,
                )
            )

    mismatchCount = gapMismatchCount + gapGapMismatchCount + nonGapMismatchCount
    append(countPrint("%sMismatches" % indent, mismatchCount, len1, len2))
    conflicts = "conflicts" if matchAmbiguous else "conflicts or ambiguities"
    append(
        countPrint(
            "%s  Not involving gaps (i.e., %s)" % (indent, conflicts),
            nonGapMismatchCount,
            len1,
            len2,
        )
    )
    append(
        countPrint(
            "%s  Involving a gap in one sequence" % indent, gapMismatchCount, len1, len2
        )
    )
    append(
        countPrint(
            "%s  Involving a gap in both sequences" % indent,
            gapGapMismatchCount,
            len1,
            len2,
        )
    )
    append(
        countPrint(
            "%s  Involving no coverage in one sequence" % indent,
            noCoverageCount,
            len1,
            len2,
        )
    )
    append(
        countPrint(
            "%s  Involving no coverage in both sequences" % indent,
            noCoverageNoCoverageCount,
            len1,
            len2,
        )
    )

    for read, key in zip((read1, read2), ("read1", "read2")):
        append("%s  Id: %s" % (indent, read.id))
        length = len(read)
        append("%s    Length: %d" % (indent, length))
        gapCount = len(dnaMatch[key]["gapOffsets"])
        append(countPrint("%s    Gaps" % indent, gapCount, length))
        if includeGapLocations and gapCount:
            append(
                "%s    Gap locations (1-based): %s"
                % (
                    indent,
                    ", ".join(
                        map(
                            lambda offset: str(offset + 1),
                            sorted(dnaMatch[key]["gapOffsets"]),
                        )
                    ),
                )
            )
        ncCount = len(dnaMatch[key]["noCoverageOffsets"])
        append(countPrint("%s    No coverage" % indent, ncCount, length))
        if includeNoCoverageLocations and ncCount:
            append(
                "%s    No coverage locations (1-based): %s"
                % (
                    indent,
                    ", ".join(
                        map(
                            lambda offset: str(offset + 1),
                            sorted(dnaMatch[key]["noCoverageOffsets"]),
                        )
                    ),
                )
            )
        ambiguousCount = len(dnaMatch[key]["ambiguousOffsets"])
        append(countPrint("%s    Ambiguous" % indent, ambiguousCount, length))
        extraCount = dnaMatch[key]["extraCount"]
        if extraCount:
            append(
                countPrint(
                    "%s    Extra nucleotides at end" % indent, extraCount, length
                )
            )

    if includeAmbiguousMatches and match["ambiguousMatches"]:
        append(f"{indent}Ambiguous matches:")
        for offset, a, b in match["ambiguousMatches"]:
            append(f"{indent}    {offset + 1} {a} {b}")

    if includeNonGapMismatches and match["nonGapMismatches"]:
        append(f"{indent}Non-gap mismatches:")
        for offset, a, b in match["nonGapMismatches"]:
            append(f"{indent}    {offset + 1} {a} {b}")

    return "\n".join(result)


def compareDNAReads(
    read1, read2, matchAmbiguous=True, gapChars="-", noCoverageChars=None, offsets=None
):
    """
    Compare two DNA sequences.

    @param read1: A C{Read} instance or an instance of one of its subclasses.
    @param read2: A C{Read} instance or an instance of one of its subclasses.
    @param matchAmbiguous: If C{True}, count ambiguous nucleotides that are
        possibly correct as actually being correct, and score these in the
        ambiguousMatchCount. Otherwise, we are strict and insist that only
        non-ambiguous nucleotides can contribute to the matching nucleotide
        count.
    @param gapChars: An object supporting __contains__ with characters that
        should be considered to be gaps.
    @param noCoverageChars: An object supporting __contains__ with characters
        that indicate sequence positions that did not have coverage (for
        comparing sequences that may be consensuses from sequencing reads
        mapped to a reference.
    @param offsets: If not C{None}, a C{set} of offsets of interest. Offsets
        not in the set will not be considered.
    @return: A C{dict} with information about the match and the individual
        sequences (see below).
    """
    identicalMatchCount = ambiguousMatchCount = 0
    gapMismatchCount = nonGapMismatchCount = gapGapMismatchCount = 0
    noCoverageCount = noCoverageNoCoverageCount = 0
    read1ExtraCount = read2ExtraCount = 0
    read1GapOffsets = []
    read2GapOffsets = []
    read1AmbiguousOffsets = []
    read2AmbiguousOffsets = []
    read1NoCoverageOffsets = []
    read2NoCoverageOffsets = []
    empty = set()
    noCoverageChars = noCoverageChars or empty
    nonGapMismatches = []
    ambiguousMatches = []

    def _identicalMatch(a, b):
        """
        Two nucleotides are considered identical if they are the same character
        and they are not ambiguous.
        """
        return a == b and AMBIGUOUS.get(a, empty) == {a}

    def _ambiguousMatch(a, b, matchAmbiguous):
        """
        Checks if two characters match ambiguously if matchAmbiguous is True.
        A match is an ambiguous match if it is not an identical match, but the
        sets of ambiguous characters overlap.
        """
        return (
            matchAmbiguous
            and not _identicalMatch(a, b)
            and AMBIGUOUS.get(a, empty) & AMBIGUOUS.get(b, empty)
        )

    for offset, (a, b) in enumerate(
        zip_longest(read1.sequence.upper(), read2.sequence.upper())
    ):
        # Use 'is not None' in the following to allow an empty offsets set
        # to be passed.
        if offsets is not None and offset not in offsets:
            continue
        if len(AMBIGUOUS.get(a, "")) > 1:
            read1AmbiguousOffsets.append(offset)
        if len(AMBIGUOUS.get(b, "")) > 1:
            read2AmbiguousOffsets.append(offset)
        aNoCoverage = a in noCoverageChars
        bNoCoverage = b in noCoverageChars
        if a is None:
            # b has an extra character at its end (it cannot be None).
            assert b is not None
            read2ExtraCount += 1
            if b in gapChars:
                read2GapOffsets.append(offset)
        elif b is None:
            # a has an extra character at its end.
            read1ExtraCount += 1
            if a in gapChars:
                read1GapOffsets.append(offset)
        elif aNoCoverage or bNoCoverage:
            if aNoCoverage and bNoCoverage:
                noCoverageNoCoverageCount += 1
                read1NoCoverageOffsets.append(offset)
                read2NoCoverageOffsets.append(offset)
            elif aNoCoverage:
                noCoverageCount += 1
                read1NoCoverageOffsets.append(offset)
            else:
                noCoverageCount += 1
                read2NoCoverageOffsets.append(offset)
        else:
            # We have a character from both sequences (they could still be
            # gap characters).
            if a in gapChars:
                read1GapOffsets.append(offset)
                if b in gapChars:
                    # Both are gaps. This could happen if the sequences are
                    # taken from a multiple sequence alignment and some
                    # other sequence has resulted in a gap being inserted
                    # in both our seqeunces. This should never happen if
                    # our sequences were pairwise aligned (no sensible
                    # alignment program would have a reason to put a gap at
                    # the same place when aligning just two sequences).
                    gapGapMismatchCount += 1
                    read2GapOffsets.append(offset)
                else:
                    # a is a gap, b is not.
                    gapMismatchCount += 1
            else:
                if b in gapChars:
                    # b is a gap, a is not.
                    gapMismatchCount += 1
                    read2GapOffsets.append(offset)
                else:
                    # Neither is a gap character.
                    if _identicalMatch(a, b):
                        identicalMatchCount += 1
                    elif _ambiguousMatch(a, b, matchAmbiguous):
                        ambiguousMatchCount += 1
                        ambiguousMatches.append((offset, a, b))
                    else:
                        nonGapMismatchCount += 1
                        nonGapMismatches.append((offset, a, b))

    return {
        "match": {
            "identicalMatchCount": identicalMatchCount,
            "ambiguousMatchCount": ambiguousMatchCount,
            "gapMismatchCount": gapMismatchCount,
            "gapGapMismatchCount": gapGapMismatchCount,
            "nonGapMismatchCount": nonGapMismatchCount,
            "noCoverageCount": noCoverageCount,
            "noCoverageNoCoverageCount": noCoverageNoCoverageCount,
            "ambiguousMatches": ambiguousMatches,
            "nonGapMismatches": nonGapMismatches,
        },
        "read1": {
            "ambiguousOffsets": read1AmbiguousOffsets,
            "extraCount": read1ExtraCount,
            "gapOffsets": read1GapOffsets,
            "noCoverageOffsets": read1NoCoverageOffsets,
        },
        "read2": {
            "ambiguousOffsets": read2AmbiguousOffsets,
            "extraCount": read2ExtraCount,
            "gapOffsets": read2GapOffsets,
            "noCoverageOffsets": read2NoCoverageOffsets,
        },
    }


def findKozakConsensus(read):
    """
    In a given DNA sequence, search for a Kozak consensus: (gcc)gccRccATGG.
    The upper case bases in that pattern are required, and the lower case
    bases are the ones most frequently found at the given positions. The
    initial 'gcc' sequence (in parentheses) is of uncertain significance
    and is not taken into account here.

    @param read: A C{DNARead} instance to be checked for Kozak consensi.
    @return: A generator that yields C{DNAKozakRead} instances.
    """
    from dark.reads import DNAKozakRead

    readLen = len(read)
    if readLen > 9:
        offset = 6
        readSeq = read.sequence
        while offset < readLen - 3:
            triplet = readSeq[offset : offset + 3]
            if triplet == "ATG":
                if readSeq[offset + 3] == "G":
                    if readSeq[offset - 3] in "GA":
                        kozakQualityCount = sum(
                            (
                                readSeq[offset - 1] == "C",
                                readSeq[offset - 2] == "C",
                                readSeq[offset - 4] == "C",
                                readSeq[offset - 5] == "C",
                                readSeq[offset - 6] == "G",
                            )
                        )

                        kozakQualityPercent = kozakQualityCount / 5.0 * 100
                        yield DNAKozakRead(
                            read, offset - 6, offset + 4, kozakQualityPercent
                        )
            offset += 1


class FloatBaseCounts:
    """
    Hold a floating point count of possible nucleotide bases.

    @param codes: An iterable of nucleotide codes.
    @param unknownAreAmbiguous: If C{True}, any unknown character (e.g., a '-'
        gap or '?' unknown base) will be treated as being fully ambiguous
        (i.e., could be any of ACGT). Otherwise, all unknown characters are
        collected under the count for '-'.
    """

    def __init__(self, codes, unknownAreAmbiguous=False):
        self.codes = list(map(str.upper, codes))
        self.unknownAreAmbiguous = unknownAreAmbiguous
        self.n = len(self.codes)
        counts = defaultdict(float)

        default = self._default = set("ACGT") if unknownAreAmbiguous else {"-"}

        for code in self.codes:
            possible = AMBIGUOUS.get(code, default)
            frac = 1.0 / len(possible)
            for p in possible:
                counts[p] += frac

        # Sort first by base.
        _sorted = [(base, counts[base]) for base in sorted(counts)]

        # Then by count (reversed).
        def key(item):
            return item[1]

        self._sorted = sorted(_sorted, key=key, reverse=True)
        self.counts = counts

    def mostFrequent(self):
        """
        Which bases are the most frequent?

        @return: A C{set} of the most frequent bases.
        """
        maxCount = self._sorted[0][1]
        return set(base for base, count in self._sorted if count == maxCount)

    def highestFrequency(self):
        """
        What is the frequency of the most frequent base?

        @return: The C{float} frequency of the most common base.
        """
        if len(self.counts) < 2:
            return 1.0
        else:
            return self._sorted[0][1] / float(self.n)

    def homogeneous(self, level):
        """
        Does the most frequent base occurs at least C{level} fraction of the
        time?

        @param level: A C{float} fraction.
        @return: C{True} if the most common base occurs at least C{level}
            fraction of the time. If there are no bases at all, this is
            considered homogeneous.
        """
        return self.highestFrequency() >= level

    def __len__(self):
        return len(self.counts)

    def __str__(self):
        fmt = "%d" if all(c == int(c) for b, c in self._sorted) else "%.2f"
        return "%s (%.3f)" % (
            " ".join(("%s:" + fmt) % (b, c) for b, c in self._sorted),
            self.highestFrequency(),
        )

    def variable(self, confirm=True):
        """
        Are the nucleotides variable?

        @param confirm: If C{True}, confirm that there is variability by
            looking at the ambiguous nucleotides. Else just report C{True}
            if there is more than one code (which might not indicate actual
            variability, since two codes could be ambiguous and have a
            nucleotide in common).
        """
        codes = self.codes

        if confirm:
            unambiguous = set()
            ambiguousIntersection = None
            default = self._default
            for code in codes:
                possible = AMBIGUOUS.get(code, default)
                if len(possible) == 1:
                    unambiguous.add(code)
                else:
                    if ambiguousIntersection is None:
                        ambiguousIntersection = set(possible)
                    else:
                        ambiguousIntersection.intersection_update(possible)

            if len(unambiguous) == 0:
                # There were no unambiguous nucleotide codes.

                # Sanity check: there must have been some ambiguous sites.
                assert ambiguousIntersection is not None

                if len(ambiguousIntersection) == 0:
                    # The ambiguous sites had nothing in common, so
                    # variation must exist (it cannot be determined what
                    # the variation is, but we don't care about that).
                    return True
                else:
                    # All the ambiguous sites have at least one nucleotide
                    # in common, so we can't be sure there's any variation.
                    pass
            elif len(unambiguous) == 1:
                # All the unambiguous sites agree. Do any of the ambiguous
                # sites (if any) not allow the unambiguous nucleotide in
                # their possibilities?
                if ambiguousIntersection is None:
                    # There were no ambiguous sites, so there's no
                    # variation here.
                    pass
                else:
                    # If any of the ambiguous sites excludes the single
                    # unambiguous nucleotide, then variation must exist.
                    nt = unambiguous.pop()
                    for code in codes:
                        possible = AMBIGUOUS.get(code, default)
                        if nt not in possible:
                            return True
            elif len(unambiguous) > 1:
                return True
        else:
            if len(codes) > 1:
                return True

        return False


def sequenceToRegex(sequence, wildcards="-?"):
    """
    Convert a potentially ambiguous DNA sequence into a regular expression.
    '?' and '-' are translated into [ACGT].

    @param sequence: A C{str} DNA sequence, possibly with ambiguous codes.
        Case insensitive.
    @param wildcards: A C{set} (or C{str}) with characters that should be
        translated to '[ACGT]'. Note that this happens only if the standard
        ambiguous lookup fails (the order could be changed one day if we need
        to override, or we could allow the passing of an ambiguity mapping).
        Wildcards are case sensitive.
    @raise KeyError: If any character in C{sequence} is unknown.
    @return: A C{str} regular expression with [...] for the ambiguous codes in
        C{sequence}.
    """
    result = []
    append = result.append
    for base in sequence.upper():
        try:
            possible = "".join(sorted(AMBIGUOUS[base]))
        except KeyError:
            if base in wildcards:
                possible = "ACGT"
            else:
                raise

        append(("[%s]" % possible) if len(possible) > 1 else possible)

    return "".join(result)


def leastAmbiguous(bases):
    """
    Get the least ambiguous code for a set of DNA bases.

    @param bases: An iterable of C{str} DNA bases.
    @raise KeyError: If C[bases} contains an unknown nucleotide.
    @return: a C{str} DNA code (one of the values of the
        BASES_TO_AMBIGUOUS dict defined at top).
    """
    return BASES_TO_AMBIGUOUS["".join(sorted(set(map(str.upper, bases))))]


def leastAmbiguousFromCounts(bases, threshold):
    """
    Get the least ambiguous code given frequency counts for a set of DNA bases
    and a homogeneity threshold.

    @param bases: A C{dict} (or C{Counter}) mapping C{str} nucleotide
        letters to C{int} frequency counts.
    @param threshold: A C{float} homogeneity frequency.
    @raise ValueError: If any count or the thrshold is less than zero.
    @return: a C{str} DNA code (one of the values of the
        BASES_TO_AMBIGUOUS dict defined at top).
    """
    total = 0
    for base, count in bases.items():
        if count < 0:
            raise ValueError(f"Count for base {base!r} is negative ({count}).")
        total += count

    if threshold < 0:
        raise ValueError(f"Threshold cannot be negative ({threshold}).")

    if total == 0:
        return leastAmbiguous("ACGT")

    counts = sorted(bases.items(), key=itemgetter(1), reverse=True)
    cumulative = 0
    resultBases = set()
    for index in range(len(counts)):
        base, count = counts[index]
        cumulative += count
        resultBases.add(base)
        if cumulative / total >= threshold:
            break

    # Add bases with counts that tie the most-recently added base.
    lastCount = count
    for index in range(index + 1, len(counts)):
        base, count = counts[index]
        if count == lastCount:
            resultBases.add(base)
        else:
            break

    return leastAmbiguous(resultBases)


class Bases:
    """
    Manage a collection of (base, quality) pairs for a genome site.
    """

    __slots__ = ("count", "counts")

    def __init__(self):
        self.count = 0
        self.counts = dict.fromkeys("ACGT", 0)

    def __eq__(self, other):
        return self.count == other.count and self.counts == other.counts

    def __str__(self):
        return f"<Bases count={self.count}, bases={self.counts}"

    __repr__ = __str__

    def __getitem__(self, key):
        return self.counts[key]

    def __len__(self):
        return self.count

    def __add__(self, other):
        new = Bases()
        new.count = self.count + other.count
        for counts in self.counts, other.counts:
            for key, count in counts.items():
                new.counts[key] += count
        return new

    def __iadd__(self, other):
        self.count += other.count
        for key, count in other.counts.items():
            self.counts[key] += count
        return self

    def append(self, base, quality):
        """
        Append a (base, quality) pair.

        @param base: A C{str} nucleotide base.
        @param quality: An C{int} nucleotide quality score.
        @return: C{self}.
        """
        if base != "N":
            self.count += 1
            self.counts[base] += quality

        return self

    def consensus(self, threshold, minCoverage, lowCoverage, noCoverage):
        """
        Get the base that can be used as part of a consensus.

        If there are sufficient reads, this is the least-ambiguous nucleotide
        code for our bases, given a required homogeneity threshold. Otherwise,
        the low coverage value.

        @param threshold: A C{float} threshold. This fraction, at least, of the
            most-common nucleotides at a site are used to determine the
            consensus nucleotide (or ambiguous symbol if more than one
            nucleotide is required to achieve this threshold). If there is a
            tie in nucleotide counts at a site that causes the threshold to be
            met, all nucleotides of equeal frequncy will be included in the
            ambiguous symbol for that site. This is perhaps better explained
            with an example. See
            https://assets.geneious.com/manual/2020.1/static/GeneiousManualse43.html
            and the corresponding testGeneiousExamplesTie test in
            test/test_dna.py
        @param minCoverage: An C{int} minimum number of reads that must cover a
            site for a consensus base to be called. If fewer reads cover a
            site, the C{lowCoverage} value is used.
        @param lowCoverage: A C{str} indicating what base to use when
            0 < N < minCoverage reads cover a site.
        @param noCoverage: A C{str} indicating what base to use when
            no reads cover the site.
        @return: A C{str} nucleotide code. This will be an ambiguous code if
            the homogeneity C{threshold} is not met.
        """
        return (
            noCoverage
            if self.count == 0
            else (
                lowCoverage
                if self.count < minCoverage
                else leastAmbiguousFromCounts(self.counts, threshold)
            )
        )



#!/usr/bin/env python

import sys
import argparse

from dark.reads import (
    addFASTACommandLineOptions,
    parseFASTACommandLineOptions,
    simpleReadSplitter,
)


parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description=(
        "Split FASTA/Q sequences longer than a given length. Print the beginning "
        "and end sequences of sufficiently long sequences separately."
    ),
)

parser.add_argument(
    "--length",
    type=int,
    required=True,
    help="Sequences longer than this will be split in two.",
)

parser.add_argument(
    "--leftPrefix",
    default="",
    help="The string to put before the id of the starting (left) subsequence.",
)

parser.add_argument(
    "--leftSuffix",
    default="",
    help="The string to put after the id of the starting (left) subsequence.",
)

parser.add_argument(
    "--rightPrefix",
    default="",
    help="The string to put before the id of the ending (right) subsequence.",
)

parser.add_argument(
    "--rightSuffix",
    default="",
    help="The string to put after the id of the ending (right) subsequence.",
)

parser.add_argument(
    "--noWarn",
    dest="warn",
    action="store_false",
    help=(
        "Do not warn the user if no prefixes are given (which can result in "
        "FASTA output with duplicated IDs."
    ),
)

addFASTACommandLineOptions(parser)
args = parser.parse_args()

assert args.length > 0, "The --length option must be positive."

if args.warn and (
    (args.leftPrefix == "" and args.leftSuffix == "")
    or (args.rightPrefix == "" and args.rightSuffix == "")
):
    print(
        "WARNING: You have not specified enough left/right ID sequence fragment "
        "prefixes/suffixes. This means found fragments (if any) will receive ids "
        "that are identical to the longer sequence they are derived from.",
        file=sys.stderr,
    )

reads = parseFASTACommandLineOptions(args)

splitter = simpleReadSplitter(
    args.length, args.leftPrefix, args.leftSuffix, args.rightPrefix, args.rightSuffix
)

for read in parseFASTACommandLineOptions(args):
    for fragment in splitter(read):
        print(fragment.toString(), end="")



import re
from operator import attrgetter
from typing import Optional
from collections import defaultdict
from collections.abc import Iterable

from dark.dna import compareDNAReads
from dark.reads import DNARead


class WindowedIdentity:
    """
    Compute windowed nucleotide identity for a set of reads against a reference.
    """

    def __init__(self, reads: Iterable[DNARead]) -> None:
        self._reads = reads
        self._validated = False

    def _checkReads(self) -> None:
        """
        Check all reads have the same length.

        @param reads: A C{list} of C{DNARead} instances.
        """
        if self._validated:
            return

        if not self._reads:
            raise ValueError("Empty FASTA input!")

        first = True

        for count, read in enumerate(self._reads, start=1):
            if first:
                length = len(read)
                if length == 0:
                    raise ValueError("The first input sequence has length zero!")
                first = False
            else:
                if len(read) != length:
                    raise ValueError(
                        f"Sequence number {count}, with id {read.id!r}, has length "
                        f"{len(read)} which is not equal to the previous input sequence "
                        f"length(s) of {length}."
                    )

        self._validated = True

    def _matchingReads(self, pattern: str, reads: Iterable[DNARead]) -> list[DNARead]:
        """
        Find the reads whose ids match a pattern. The matching looks for the regular
        expression anywhere in the read ids.

        @param pattern: A regular expression.
        @param reads: The reads to examine.
        @return: A list of matching reads.
        """
        regex = re.compile(pattern)
        return [read for read in reads if regex.search(read.id)]

    def getIdentity(
        self,
        referenceRegex: str,
        window: int,
        minWindow: int,
        jump: int,
        includeRegex: str = ".",
        sort: bool = False,
        strict: bool = True,
        startOffset: int = 0,
        stopOffset: Optional[int] = None,
    ) -> tuple[DNARead, list[int], dict[DNARead, list[float]]]:
        """
        Compute windowed mucleotide identity of the reads whose ids are matched by
        includeRegex against the read whose id is matched by referenceRegex.

        @param referenceRegex: A C{str} regular expression indicating the reference
            sequence id (i.e., the sequence that others should be compared to. This must
            obviously match exactly one input sequence.
        @param window: The C{int} window size.
        @param minWindow: The C{int} minimum window size.
        @param jump: The C{int} amount to move along the genome to set up the next window.
        @param includeRegex: A regular expression indicating the sequence ids to include
            in the plot. If not specified, all sequences (other than the reference
            itself) will be compared to the reference.
        @param sort: Sort the ids of the sequences that are compared. If false, the
            order that matching ids are found in the input FASTA will be used.
        @param strict: If true, matching will be strict (ambiguous nucleotide codes
            will be considered not to match).
        @param startOffset: The C{int} zero-based genome offset to start the first
            window at.
        @param stopOffset: The C{int} zero-based genome offset to stop the last
            window at.
        @return: A C{tuple} containing the reference C{DNARead}, a C{list} of C{int}
            window start offsets, and a C{dict} keyed by C{DNARead} (for the reads whose
            ids are matched by C{includeRegex}), with values being a C{list} of C{float}
            nucleotide identity values for the windows across the genome.
        """
        self._checkReads()

        if window < 1:
            raise ValueError(f"Window size must be positive. You passed {window}.")

        if minWindow < 0:
            raise ValueError(
                f"Minimum window size cannot be negative. You passed {minWindow}."
            )

        if jump < 1:
            raise ValueError(f"Jump must be positive. You passed {jump}.")

        if startOffset < 0:
            raise ValueError(
                f"Start offset cannot be negative. You passed {startOffset}."
            )

        if stopOffset is not None:
            if stopOffset < 0:
                raise ValueError(
                    f"Stop offset cannot be negative. You passed {stopOffset}."
                )
            if stopOffset <= startOffset:
                raise ValueError(
                    f"Stop offset must be greater than start offset. You passed "
                    f"{startOffset = } and {stopOffset = }."
                )

        references = list(self._matchingReads(referenceRegex, self._reads))
        nReferences = len(references)

        # We can't use 'match' until we've dropped support for Python 3.9.
        if nReferences == 0:
            raise ValueError(
                f"No input sequence IDs match the regular expression for the reference "
                f"({referenceRegex!r})."
            )
        elif nReferences == 1:
            reference = references[0]
        else:
            ids = ", ".join(sorted(f"{read.id!r}" for read in references))
            raise ValueError(
                f"{len(references)} input sequence ids match the reference "
                f"pattern! The matching ids are: {ids}."
            )

        toCompare = list(
            set(self._matchingReads(includeRegex, self._reads)) - {reference}
        )
        nToCompare = len(toCompare)

        if nToCompare == 0:
            raise ValueError("No input sequence ids match the include pattern!")

        if sort:
            toCompare.sort(key=attrgetter("id"))

        if stopOffset is None:
            length = len(reference)
        else:
            length = min(len(reference), stopOffset)

        identity = defaultdict(list)
        starts = []
        start = startOffset

        while True:
            stop = start + window
            if stop > length:
                stop = length

            thisLength = stop - start
            if thisLength < minWindow:
                break

            starts.append(start)

            for read in toCompare:
                match = compareDNAReads(
                    reference[start:stop], read[start:stop], matchAmbiguous=not strict
                )["match"]

                identity[read].append(
                    (match["identicalMatchCount"] + match["ambiguousMatchCount"])
                    / thisLength
                )

            start += jump

        return reference, starts, identity


def addCommandLineOptions(parser):
    """
    Add options to an argument parser for any command-line tool that wants to
    work with windowed sequences.

    Note that there might be naming collisions due to the use below of simple
    option names, like --sort.
    """

    parser.add_argument(
        "--reference",
        required=True,
        help=(
            "A regular expression indicating the reference sequence id (i.e., the "
            "sequence that others should be compared to. This must obviously match "
            "exactly one input sequence."
        ),
    )

    parser.add_argument(
        "--include",
        default=".",
        help=(
            "A regular expression indicating the sequence ids to include in the plot. "
            "If not specified, all sequences (other than the reference itself) will "
            "be compared to the reference."
        ),
    )

    parser.add_argument(
        "--window",
        default=200,
        type=int,
        help="The width (number of nucleotides) of the window.",
    )

    parser.add_argument(
        "--minWindow",
        default=50,
        type=int,
        help="The smallest window size to examine.",
    )

    parser.add_argument(
        "--jump",
        default=50,
        type=int,
        help=(
            "The number of nucleotides to move across to the start of the "
            "next window."
        ),
    )

    parser.add_argument(
        "--sort",
        action="store_true",
        help=(
            "Sort the ids of the sequences that are compared. If not given, the "
            "order that matching ids are found in the input FASTA will be used."
        ),
    )

    parser.add_argument(
        "--startOffset",
        default=0,
        type=int,
        help="The zero-based genome offset to start the first window at.",
    )

    parser.add_argument(
        "--stopOffset",
        type=int,
        help="The zero-based genome offset to stop the last window at.",
    )



from unittest import TestCase

from dark.idutils import ids


class TestIds(TestCase):
    """
    Test the ids function.
    """

    def testStartAndStopStringsOfUnequalLength(self):
        """
        If the passed start and end string are not of the same length, a ValueError
        must be raised.
        """
        error = r"^Range specifiers must be of equal length\.$"
        self.assertRaisesRegex(ValueError, error, tuple, ids("xxx", "yyyy"))

    def testNothing(self):
        """
        If no ids are requested, an empty list must result.
        """
        self.assertEqual((), tuple(ids("Pool-AB", maxResults=0)))

    def testMaxCount(self):
        """
        A non-zero maximum count must work as expected with just a start
        value passed.
        """
        self.assertEqual(("aa", "ab", "ac", "ad"), tuple(ids("aa", maxResults=4)))

    def testMaxCountCutsUnfinishedIds(self):
        """
        A non-zero maximum count must work as expected when the number of ids
        that would be returned exceeds the count.
        """
        self.assertEqual(("6", "7", "8"), tuple(ids(6, 9, maxResults=3)))

    def testTwoNakedIntsAscending(self):
        """
        Test two naked integers in ascending order.
        """
        self.assertEqual(("5", "6", "7"), tuple(ids(5, 7)))

    def testTwoNakedIntsDescending(self):
        """
        Test two naked integers in descending order.
        """
        self.assertEqual(("5", "6", "7"), tuple(ids(7, 5)))

    def testTwoIntsInAStringAscending(self):
        """
        Test two integers in ascending order in a string.
        """
        self.assertEqual(("id-5", "id-6", "id-7"), tuple(ids("id-5", "id-7")))

    def testTwoIntsInAStringDescending(self):
        """
        Test two integers given in descending order in a string.
        """
        self.assertEqual(("id-5", "id-6", "id-7"), tuple(ids("id-7", "id-5")))

    def testTwoIntsAscendingExplicitPrefixPassingStrings(self):
        """
        Test two integers in ascending order when arguments are passed as strings
        and an explicit prefix is given.
        """
        self.assertEqual(("id-5", "id-6", "id-7"), tuple(ids("5", "7", prefix="id-")))

    def testTwoIntsAscendingExplicitPrefixPassingInts(self):
        """
        Test two integers in ascending order when arguments are passed as ints
        and an explicit prefix is given.
        """
        self.assertEqual(("id-5", "id-6", "id-7"), tuple(ids(5, 7, prefix="id-")))

    def testTwoIntsDescendingExplicitPrefix(self):
        """
        Test two integers given in descending order when an explicit prefix is given.
        """
        self.assertEqual(("id-5", "id-6", "id-7"), tuple(ids(7, 5, prefix="id-")))

    def testIntermediateSymbols(self):
        """
        It must be possible to put intermediate (constant) symbols into ids.
        """
        self.assertEqual(
            ("id-5-7", "id-5-8", "id-5-9", "id-6-0", "id-6-1"),
            tuple(ids("id-5-7", "id-6-1")),
        )

    def testIntsWithNoLimit(self):
        """
        A starting int with no limit must iterate to the next power of ten.
        """
        self.assertEqual(("93", "94", "95", "96", "97", "98", "99"), tuple(ids(93)))

    def testTwoLetterPoolIdsWithStartAndNoStop(self):
        """
        Check that example fictitional pool ids work as expected when a start is
        given but a stop is not.
        """
        self.assertEqual(
            ("Pool-ZW", "Pool-ZX", "Pool-ZY", "Pool-ZZ"),
            tuple(ids("ZW", prefix="Pool-")),
        )

    def testTwoLetterPoolIdsWithStartAndStop(self):
        """
        Check that example fictitional pool ids work as expected when a start and
        stop id are both given.
        """
        self.assertEqual(
            ("Pool-BW", "Pool-BX", "Pool-BY", "Pool-BZ", "Pool-CA", "Pool-CB"),
            tuple(ids("Pool-BW", "Pool-CB")),
        )

    def testRandomThreeLetterStringsWithNoPrefix(self):
        """
        Check that an arbitrary series of three-letter strings can be generated.
        """
        self.assertEqual(
            (
                "hnx",
                "hny",
                "hnz",
                "hoa",
                "hob",
                "hoc",
                "hod",
                "hoe",
                "hof",
                "hog",
                "hoh",
                "hoi",
                "hoj",
                "hok",
                "hol",
                "hom",
                "hon",
                "hoo",
                "hop",
            ),
            tuple(ids("hnx", "hop", prefix="")),
        )



import sys

if sys.version_info < (3, 9):
    raise Exception("The dark matter code needs Python 3.9 or later.")

# Note that the version string below must have the following format,
# otherwise it will not be found by the version() function in ../setup.py
#
# Remember to update ../CHANGELOG.md describing what's new in each version.
__version__ = "5.0.14"



#!/usr/bin/env python

import sys
import argparse
from json import dump

from dark.sam import DistanceMatrix


def main(args):
    """
    Print a JSON object with reference to read scores.

    @param args: A C{Namespace} object returned by argparse, containing
        argument values passed on the command line.
    """
    dm = DistanceMatrix()
    for filename in args.samFile:
        dm.addFile(filename, scoreTag=args.scoreTag)

    if args.minMatchingReads is None:
        wantedReferenceIds = sorted(set(dm.scores))
    else:
        wantedReferenceIds = sorted(
            set(
                referenceId
                for (referenceId, reads) in dm.scores.items()
                if len(reads) >= args.minMatchingReads
            )
        )
        if args.verbose:
            print(
                f"Found {len(wantedReferenceIds)} references with at least "
                f"{args.minMatchingReads} matching reads.",
                file=sys.stderr,
            )

    if args.verbose:
        for referenceId in wantedReferenceIds:
            nReads = len(dm.scores[referenceId])
            print(f"Reference {referenceId!r} matched {nReads} reads.")

    dump(
        dict((id_, dm.scores[id_]) for id_ in wantedReferenceIds),
        sys.stdout,
        sort_keys=True,
        indent=4,
    )
    print()


def makeParser():
    """
    Make a command-line argument parser.

    @return: An C{argparse.ArgumentParser} instance.
    """
    parser = argparse.ArgumentParser(
        description=(
            "Print a JSON object containing reference to read "
            "distances extracted from a SAM file."
        )
    )

    parser.add_argument(
        "--samFile",
        action="append",
        required=True,
        help="The SAM file(s) to load. May be repeated.",
    )

    parser.add_argument(
        "--minMatchingReads",
        type=int,
        help=(
            "The minimum number of reads that must match a reference for it "
            "to be included."
        ),
    )

    parser.add_argument(
        "--scoreTag",
        help=(
            "The score tag to use for the alignment score. If not given, "
            "1 will be used to indicate that a read matched a reference "
            "(non-matches are not included). The default is no score tag, "
            'which is not that useful. A good choice is "AS", for the '
            "alignment score, but that has to be present in the SAM file, "
            "which means that the aligner (bowtie2, bwa, etc. has to have "
            "produced such a tag."
        ),
    )

    parser.add_argument(
        "--verbose", action="store_true", help="Print extra information."
    )

    return parser


if __name__ == "__main__":
    main(makeParser().parse_args())



# From https://en.wikipedia.org/wiki/Amino_acid
#
# Alanine          Ala     A
# Arginine         Arg     R
# Asparagine       Asn     N
# Aspartic acid    Asp     D
# Cysteine         Cys     C
# Glutamic acid    Glu     E
# Glutamine        Gln     Q
# Glycine          Gly     G
# Histidine        His     H
# Isoleucine       Ile     I
# Leucine          Leu     L
# Lysine           Lys     K
# Methionine       Met     M
# Phenylalanine    Phe     F
# Proline          Pro     P
# Serine           Ser     S
# Threonine        Thr     T
# Tryptophan       Trp     W
# Tyrosine         Tyr     Y
# Valine           Val     V

NAMES: dict[str, str] = {
    "A": "Alanine",
    "R": "Arginine",
    "N": "Asparagine",
    "D": "Aspartic acid",
    "C": "Cysteine",
    "E": "Glutamic acid",
    "Q": "Glutamine",
    "G": "Glycine",
    "H": "Histidine",
    "I": "Isoleucine",
    "L": "Leucine",
    "K": "Lysine",
    "M": "Methionine",
    "F": "Phenylalanine",
    "P": "Proline",
    "S": "Serine",
    "T": "Threonine",
    "V": "Valine",
    "W": "Tryptophan",
    "Y": "Tyrosine",
}

AA_LETTERS: list[str] = list(sorted(NAMES))

NAMES_TO_ABBREV1 = dict((name, abbrev1) for abbrev1, name in NAMES.items())

ABBREV3 = {
    "A": "Ala",
    "R": "Arg",
    "N": "Asn",
    "D": "Asp",
    "C": "Cys",
    "E": "Glu",
    "Q": "Gln",
    "G": "Gly",
    "H": "His",
    "I": "Ile",
    "L": "Leu",
    "K": "Lys",
    "M": "Met",
    "F": "Phe",
    "P": "Pro",
    "S": "Ser",
    "T": "Thr",
    "V": "Val",
    "W": "Trp",
    "Y": "Tyr",
}

ABBREV3_TO_ABBREV1 = dict((abbrev3, abbrev1) for abbrev1, abbrev3 in ABBREV3.items())

HYDROPHOBIC = 0x0001
HYDROPHILIC = 0x0002
AROMATIC = 0x0004
SULPHUR = 0x0008
ALIPHATIC = 0x0010
HYDROXYLIC = 0x0020
TINY = 0x0040
SMALL = 0x0080
ACIDIC = 0x0100
BASIC_POSITIVE = 0x0200
NEGATIVE = 0x0400
POLAR = 0x0800
NONE = 0x1000


ALL_PROPERTIES = (
    ACIDIC,
    ALIPHATIC,
    AROMATIC,
    BASIC_POSITIVE,
    HYDROPHILIC,
    HYDROPHOBIC,
    HYDROXYLIC,
    NEGATIVE,
    NONE,
    POLAR,
    SMALL,
    SULPHUR,
    TINY,
)

PROPERTY_NAMES = {
    ACIDIC: "Acidic",
    ALIPHATIC: "Aliphatic",
    AROMATIC: "Aromatic",
    BASIC_POSITIVE: "Basic positive",
    HYDROPHILIC: "Hydrophilic",
    HYDROPHOBIC: "Hydrophobic",
    HYDROXYLIC: "Hydroxylic",
    NEGATIVE: "Negative",
    NONE: "<NONE>",
    POLAR: "Polar",
    SMALL: "Small",
    SULPHUR: "Sulphur",
    TINY: "Tiny",
}

PROPERTIES = {
    "A": HYDROPHOBIC | SMALL | TINY,
    "C": HYDROPHOBIC | SMALL | TINY | SULPHUR,
    "D": HYDROPHILIC | SMALL | POLAR | NEGATIVE,
    "E": HYDROPHILIC | NEGATIVE | ACIDIC,
    "F": HYDROPHOBIC | AROMATIC,
    "G": HYDROPHILIC | SMALL | TINY,
    "H": HYDROPHOBIC | AROMATIC | POLAR | BASIC_POSITIVE,
    "I": ALIPHATIC | HYDROPHOBIC,
    "K": HYDROPHOBIC | BASIC_POSITIVE | POLAR,
    "L": ALIPHATIC | HYDROPHOBIC,
    "M": HYDROPHOBIC | SULPHUR,
    "N": HYDROPHILIC | SMALL | POLAR | ACIDIC,
    "P": HYDROPHILIC | SMALL,
    "Q": HYDROPHILIC | POLAR | ACIDIC,
    "R": HYDROPHILIC | POLAR | BASIC_POSITIVE,
    "S": HYDROPHILIC | SMALL | POLAR | HYDROXYLIC,
    "T": HYDROPHOBIC | SMALL | HYDROXYLIC,
    "V": ALIPHATIC | HYDROPHOBIC | SMALL,
    "W": HYDROPHOBIC | AROMATIC | POLAR,
    "Y": HYDROPHOBIC | AROMATIC | POLAR,
}


# A table with which codons translate to which amino acids.
# Based on https://en.wikipedia.org/wiki/DNA_codon_table
#
# Note that the trailing commas are necessary in the AAs that only have one
# codon. If you omit them, the parentheses will not create a tuple.

CODONS = {
    "A": (
        "GCA",
        "GCC",
        "GCG",
        "GCT",
    ),
    "C": (
        "TGC",
        "TGT",
    ),
    "D": (
        "GAC",
        "GAT",
    ),
    "E": (
        "GAA",
        "GAG",
    ),
    "F": (
        "TTC",
        "TTT",
    ),
    "G": (
        "GGA",
        "GGC",
        "GGG",
        "GGT",
    ),
    "H": (
        "CAC",
        "CAT",
    ),
    "I": (
        "ATA",
        "ATC",
        "ATT",
    ),
    "K": (
        "AAA",
        "AAG",
    ),
    "L": (
        "CTA",
        "CTC",
        "CTG",
        "CTT",
        "TTA",
        "TTG",
    ),
    "M": ("ATG",),
    "N": (
        "AAC",
        "AAT",
    ),
    "P": (
        "CCA",
        "CCC",
        "CCG",
        "CCT",
    ),
    "Q": (
        "CAA",
        "CAG",
    ),
    "R": (
        "AGA",
        "AGG",
        "CGA",
        "CGC",
        "CGG",
        "CGT",
    ),
    "S": (
        "AGC",
        "AGT",
        "TCA",
        "TCC",
        "TCG",
        "TCT",
    ),
    "T": (
        "ACA",
        "ACC",
        "ACG",
        "ACT",
    ),
    "V": (
        "GTA",
        "GTC",
        "GTG",
        "GTT",
    ),
    "W": ("TGG",),
    "Y": (
        "TAC",
        "TAT",
    ),
}

REVERSE_CODONS = dict((codon, aa) for aa, codons in CODONS.items() for codon in codons)

START_CODON = "ATG"
STOP_CODONS = (
    "TAA",
    "TAG",
    "TGA",
)

"""
The dictionary below contains for each amino acid the value for
each property scaled from -1 to 1.
For documentation, check https://notebooks.antigenic-cartography.org/barbara/
pages/features/aa-properties.html

The following references were used for the different aa properties:

Aliphaticity, aromaticity, hydrogenation, hydroxyethilation: Sneath,
1966, Journal of Theoretical Biology, Relations between Chemical Structure
and Biological Activity in Peptides

Composition, polarity, volume: Grantham 1974, Science, Amino Acid Difference
Formula to Help Explain ProteinEvolution

Hydropathy: Kyte and Doolittle, 1982, Journal of Molecular Biology, A Simple
Method for Displaying the Hydropathic Character of a Protein

IEP: Yu et al., 2011, Gene, Protein map: An alignment-free sequence comparison
method based on various properties of amino acids

Polar requirement: Mathew and Luthey-Schulten, 2008, Journal of Molecular
Evolution, On the Physical Basis of the Amino Acid Polar Requirement
"""

PROPERTY_DETAILS = {
    "A": {
        "aliphaticity": 0.305785123967,
        "aromaticity": -0.550128534704,
        "composition": -1.0,
        "hydrogenation": 0.8973042362,
        "hydropathy": 0.4,
        "hydroxythiolation": -0.265160523187,
        "iep": -0.191489361702,
        "polar requirement": -0.463414634146,
        "polarity": -0.20987654321,
        "volume": -0.664670658683,
    },
    "C": {
        "aliphaticity": -0.00826446280992,
        "aromaticity": -0.740359897172,
        "composition": 1.0,
        "hydrogenation": 0.240051347882,
        "hydropathy": 0.555555555556,
        "hydroxythiolation": 0.785969084423,
        "iep": -0.424280350438,
        "polar requirement": -1.0,
        "polarity": -0.851851851852,
        "volume": -0.377245508982,
    },
    "D": {
        "aliphaticity": -0.818181818182,
        "aromaticity": -1.0,
        "composition": 0.00363636363636,
        "hydrogenation": -0.90243902439,
        "hydropathy": -0.777777777778,
        "hydroxythiolation": -0.348394768133,
        "iep": -1.0,
        "polar requirement": 1.0,
        "polarity": 1.0,
        "volume": -0.389221556886,
    },
    "E": {
        "aliphaticity": -0.553719008264,
        "aromaticity": -0.899742930591,
        "composition": -0.330909090909,
        "hydrogenation": -1.0,
        "hydropathy": -0.777777777778,
        "hydroxythiolation": -0.555291319857,
        "iep": -0.887359198999,
        "polar requirement": 0.878048780488,
        "polarity": 0.827160493827,
        "volume": -0.0419161676647,
    },
    "F": {
        "aliphaticity": 0.223140495868,
        "aromaticity": 0.858611825193,
        "composition": -1.0,
        "hydrogenation": 0.0218228498074,
        "hydropathy": 0.622222222222,
        "hydroxythiolation": 0.0582639714625,
        "iep": -0.321652065081,
        "polar requirement": -0.951219512195,
        "polarity": -0.925925925926,
        "volume": 0.544910179641,
    },
    "G": {
        "aliphaticity": -1.0,
        "aromaticity": -0.45501285347,
        "composition": -0.461818181818,
        "hydrogenation": 1.0,
        "hydropathy": -0.0888888888889,
        "hydroxythiolation": -0.158145065398,
        "iep": -0.198998748436,
        "polar requirement": -0.243902439024,
        "polarity": 0.0123456790123,
        "volume": -1.0,
    },
    "H": {
        "aliphaticity": -0.256198347107,
        "aromaticity": 0.555269922879,
        "composition": -0.578181818182,
        "hydrogenation": -0.150192554557,
        "hydropathy": -0.711111111111,
        "hydroxythiolation": 0.0154577883472,
        "iep": 0.206508135169,
        "polar requirement": -0.121951219512,
        "polarity": 0.358024691358,
        "volume": 0.11377245509,
    },
    "I": {
        "aliphaticity": 0.867768595041,
        "aromaticity": -0.264781491003,
        "composition": -1.0,
        "hydrogenation": 0.432605905006,
        "hydropathy": 1.0,
        "hydroxythiolation": -0.85255648038,
        "iep": -0.18648310388,
        "polar requirement": -0.975609756098,
        "polarity": -0.925925925926,
        "volume": 0.293413173653,
    },
    "K": {
        "aliphaticity": 0.123966942149,
        "aromaticity": -0.141388174807,
        "composition": -0.76,
        "hydrogenation": -0.142490372272,
        "hydropathy": -0.866666666667,
        "hydroxythiolation": -1.0,
        "iep": 0.744680851064,
        "polar requirement": 0.292682926829,
        "polarity": 0.58024691358,
        "volume": 0.389221556886,
    },
    "L": {
        "aliphaticity": 1.0,
        "aromaticity": -0.287917737789,
        "composition": -1.0,
        "hydrogenation": 0.381258023107,
        "hydropathy": 0.844444444444,
        "hydroxythiolation": -0.745541022592,
        "iep": -0.196495619524,
        "polar requirement": -0.975609756098,
        "polarity": -1.0,
        "volume": 0.293413173653,
    },
    "M": {
        "aliphaticity": 0.537190082645,
        "aromaticity": -0.372750642674,
        "composition": -1.0,
        "hydrogenation": -0.186136071887,
        "hydropathy": 0.422222222222,
        "hydroxythiolation": 0.0653983353151,
        "iep": -0.256570713392,
        "polar requirement": -0.878048780488,
        "polarity": -0.802469135802,
        "volume": 0.221556886228,
    },
    "N": {
        "aliphaticity": 0.471074380165,
        "aromaticity": -0.616966580977,
        "composition": -0.0327272727273,
        "hydrogenation": -0.548138639281,
        "hydropathy": -0.777777777778,
        "hydroxythiolation": 0.277051129608,
        "iep": -0.339173967459,
        "polar requirement": 0.268292682927,
        "polarity": 0.654320987654,
        "volume": -0.365269461078,
    },
    "P": {
        "aliphaticity": -0.917355371901,
        "aromaticity": -0.308483290488,
        "composition": -0.716363636364,
        "hydrogenation": 1.0,
        "hydropathy": -0.355555555556,
        "hydroxythiolation": -0.203329369798,
        "iep": -0.116395494368,
        "polar requirement": -0.560975609756,
        "polarity": -0.234567901235,
        "volume": -0.646706586826,
    },
    "Q": {
        "aliphaticity": 0.652892561983,
        "aromaticity": -0.439588688946,
        "composition": -0.352727272727,
        "hydrogenation": -0.602053915276,
        "hydropathy": -0.777777777778,
        "hydroxythiolation": -0.177170035672,
        "iep": -0.279098873592,
        "polar requirement": -0.0731707317073,
        "polarity": 0.382716049383,
        "volume": -0.0179640718563,
    },
    "R": {
        "aliphaticity": -0.157024793388,
        "aromaticity": -0.0642673521851,
        "composition": -0.527272727273,
        "hydrogenation": -0.401797175866,
        "hydropathy": -1.0,
        "hydroxythiolation": -0.51486325802,
        "iep": 1.0,
        "polar requirement": 0.0487804878049,
        "polarity": 0.382716049383,
        "volume": 0.449101796407,
    },
    "S": {
        "aliphaticity": 0.256198347107,
        "aromaticity": -0.660668380463,
        "composition": 0.0327272727273,
        "hydrogenation": 0.106546854942,
        "hydropathy": -0.177777777778,
        "hydroxythiolation": 1.0,
        "iep": -0.271589486859,
        "polar requirement": -0.341463414634,
        "polarity": 0.0617283950617,
        "volume": -0.652694610778,
    },
    "T": {
        "aliphaticity": -0.123966942149,
        "aromaticity": -0.80205655527,
        "composition": -0.483636363636,
        "hydrogenation": 0.399229781772,
        "hydropathy": -0.155555555556,
        "hydroxythiolation": 0.709869203329,
        "iep": -0.151439299124,
        "polar requirement": -0.560975609756,
        "polarity": -0.0864197530864,
        "volume": -0.305389221557,
    },
    "V": {
        "aliphaticity": 0.570247933884,
        "aromaticity": -0.665809768638,
        "composition": -1.0,
        "hydrogenation": 0.679075738126,
        "hydropathy": 0.933333333333,
        "hydroxythiolation": -0.621878715815,
        "iep": -0.201501877347,
        "polar requirement": -0.80487804878,
        "polarity": -0.753086419753,
        "volume": -0.0299401197605,
    },
    "W": {
        "aliphaticity": -0.619834710744,
        "aromaticity": 1.0,
        "composition": -0.905454545455,
        "hydrogenation": 0.0218228498074,
        "hydropathy": -0.2,
        "hydroxythiolation": 0.00118906064209,
        "iep": -0.219023779725,
        "polar requirement": -0.90243902439,
        "polarity": -0.876543209877,
        "volume": 1.0,
    },
    "Y": {
        "aliphaticity": -0.454545454545,
        "aromaticity": 0.712082262211,
        "composition": -0.854545454545,
        "hydrogenation": -0.304236200257,
        "hydropathy": 0.288888888889,
        "hydroxythiolation": 0.405469678954,
        "iep": -0.276595744681,
        "polar requirement": -0.853658536585,
        "polarity": -0.679012345679,
        "volume": 0.592814371257,
    },
}


"""
The dictionary below contains for each amino acid the value for
each property.
"""

PROPERTY_DETAILS_RAW = {
    "A": {
        "aliphaticity": 0.239,
        "aromaticity": -0.11,
        "composition": 0.0,
        "hydrogenation": 0.33,
        "hydropathy": 1.8,
        "hydroxythiolation": -0.062,
        "iep": 6.0,
        "polar requirement": 7.0,
        "polarity": 8.1,
        "volume": 31.0,
    },
    "C": {
        "aliphaticity": 0.22,
        "aromaticity": -0.184,
        "composition": 2.75,
        "hydrogenation": 0.074,
        "hydropathy": 2.5,
        "hydroxythiolation": 0.38,
        "iep": 5.07,
        "polar requirement": 4.8,
        "polarity": 5.5,
        "volume": 55.0,
    },
    "D": {
        "aliphaticity": 0.171,
        "aromaticity": -0.285,
        "composition": 1.38,
        "hydrogenation": -0.371,
        "hydropathy": -3.5,
        "hydroxythiolation": -0.079,
        "iep": 2.77,
        "polar requirement": 13.0,
        "polarity": 13.0,
        "volume": 54.0,
    },
    "E": {
        "aliphaticity": 0.187,
        "aromaticity": -0.246,
        "composition": 0.92,
        "hydrogenation": -0.409,
        "hydropathy": -3.5,
        "hydroxythiolation": -0.184,
        "iep": 3.22,
        "polar requirement": 12.5,
        "polarity": 12.3,
        "volume": 83.0,
    },
    "F": {
        "aliphaticity": 0.234,
        "aromaticity": 0.438,
        "composition": 0.0,
        "hydrogenation": -0.011,
        "hydropathy": 2.8,
        "hydroxythiolation": 0.074,
        "iep": 5.48,
        "polar requirement": 5.0,
        "polarity": 5.4,
        "volume": 132.0,
    },
    "G": {
        "aliphaticity": 0.16,
        "aromaticity": -0.073,
        "composition": 0.74,
        "hydrogenation": 0.37,
        "hydropathy": -0.4,
        "hydroxythiolation": -0.017,
        "iep": 5.97,
        "polar requirement": 7.9,
        "polarity": 9.0,
        "volume": 3.0,
    },
    "H": {
        "aliphaticity": 0.205,
        "aromaticity": 0.32,
        "composition": 0.58,
        "hydrogenation": -0.078,
        "hydropathy": -3.2,
        "hydroxythiolation": 0.056,
        "iep": 7.59,
        "polar requirement": 8.4,
        "polarity": 10.4,
        "volume": 96.0,
    },
    "I": {
        "aliphaticity": 0.273,
        "aromaticity": 0.001,
        "composition": 0.0,
        "hydrogenation": 0.149,
        "hydropathy": 4.5,
        "hydroxythiolation": -0.309,
        "iep": 6.02,
        "polar requirement": 4.9,
        "polarity": 5.2,
        "volume": 111.0,
    },
    "K": {
        "aliphaticity": 0.228,
        "aromaticity": 0.049,
        "composition": 0.33,
        "hydrogenation": -0.075,
        "hydropathy": -3.9,
        "hydroxythiolation": -0.371,
        "iep": 9.74,
        "polar requirement": 10.1,
        "polarity": 11.3,
        "volume": 119.0,
    },
    "L": {
        "aliphaticity": 0.281,
        "aromaticity": -0.008,
        "composition": 0.0,
        "hydrogenation": 0.129,
        "hydropathy": 3.8,
        "hydroxythiolation": -0.264,
        "iep": 5.98,
        "polar requirement": 4.9,
        "polarity": 4.9,
        "volume": 111.0,
    },
    "M": {
        "aliphaticity": 0.253,
        "aromaticity": -0.041,
        "composition": 0.0,
        "hydrogenation": -0.092,
        "hydropathy": 1.9,
        "hydroxythiolation": 0.077,
        "iep": 5.74,
        "polar requirement": 5.3,
        "polarity": 5.7,
        "volume": 105.0,
    },
    "N": {
        "aliphaticity": 0.249,
        "aromaticity": -0.136,
        "composition": 1.33,
        "hydrogenation": -0.233,
        "hydropathy": -3.5,
        "hydroxythiolation": 0.166,
        "iep": 5.41,
        "polar requirement": 10.0,
        "polarity": 11.6,
        "volume": 56.0,
    },
    "P": {
        "aliphaticity": 0.165,
        "aromaticity": -0.016,
        "composition": 0.39,
        "hydrogenation": 0.37,
        "hydropathy": -1.6,
        "hydroxythiolation": -0.036,
        "iep": 6.3,
        "polar requirement": 6.6,
        "polarity": 8.0,
        "volume": 32.5,
    },
    "Q": {
        "aliphaticity": 0.26,
        "aromaticity": -0.067,
        "composition": 0.89,
        "hydrogenation": -0.254,
        "hydropathy": -3.5,
        "hydroxythiolation": -0.025,
        "iep": 5.65,
        "polar requirement": 8.6,
        "polarity": 10.5,
        "volume": 85.0,
    },
    "R": {
        "aliphaticity": 0.211,
        "aromaticity": 0.079,
        "composition": 0.65,
        "hydrogenation": -0.176,
        "hydropathy": -4.5,
        "hydroxythiolation": -0.167,
        "iep": 10.76,
        "polar requirement": 9.1,
        "polarity": 10.5,
        "volume": 124.0,
    },
    "S": {
        "aliphaticity": 0.236,
        "aromaticity": -0.153,
        "composition": 1.42,
        "hydrogenation": 0.022,
        "hydropathy": -0.8,
        "hydroxythiolation": 0.47,
        "iep": 5.68,
        "polar requirement": 7.5,
        "polarity": 9.2,
        "volume": 32.0,
    },
    "T": {
        "aliphaticity": 0.213,
        "aromaticity": -0.208,
        "composition": 0.71,
        "hydrogenation": 0.136,
        "hydropathy": -1.3,
        "hydroxythiolation": 0.348,
        "iep": 6.16,
        "polar requirement": 6.6,
        "polarity": 8.6,
        "volume": 61.0,
    },
    "V": {
        "aliphaticity": 0.255,
        "aromaticity": -0.155,
        "composition": 0.0,
        "hydrogenation": 0.245,
        "hydropathy": 4.2,
        "hydroxythiolation": -0.212,
        "iep": 5.96,
        "polar requirement": 5.6,
        "polarity": 5.9,
        "volume": 84.0,
    },
    "W": {
        "aliphaticity": 0.183,
        "aromaticity": 0.493,
        "composition": 0.13,
        "hydrogenation": -0.011,
        "hydropathy": -0.9,
        "hydroxythiolation": 0.05,
        "iep": 5.89,
        "polar requirement": 5.2,
        "polarity": 5.4,
        "volume": 170.0,
    },
    "Y": {
        "aliphaticity": 0.193,
        "aromaticity": 0.183,
        "composition": 0.2,
        "hydrogenation": -0.138,
        "hydropathy": -1.3,
        "hydroxythiolation": 0.22,
        "iep": 5.66,
        "polar requirement": 5.4,
        "polarity": 6.2,
        "volume": 136.0,
    },
}


"""
Clusters based on raw amino acid property values. See
https://notebooks.antigenic-cartography.org/barbara/pages/features/
aa-properties.html and https://notebooks.antigenic-cartography.org/barbara/
pages/features/new-tps.html
"""

PROPERTY_CLUSTERS = {
    "A": {
        "aliphaticity": 1,
        "aromaticity": 1,
        "composition": 1,
        "hydrogenation": 1,
        "hydropathy": 3,
        "hydroxythiolation": 2,
        "iep": 2,
        "polar requirement": 2,
        "polarity": 2,
        "volume": 2,
    },
    "C": {
        "aliphaticity": 1,
        "aromaticity": 1,
        "composition": 3,
        "hydrogenation": 1,
        "hydropathy": 3,
        "hydroxythiolation": 5,
        "iep": 2,
        "polar requirement": 1,
        "polarity": 1,
        "volume": 3,
    },
    "D": {
        "aliphaticity": 1,
        "aromaticity": 1,
        "composition": 2,
        "hydrogenation": 1,
        "hydropathy": 1,
        "hydroxythiolation": 2,
        "iep": 1,
        "polar requirement": 4,
        "polarity": 4,
        "volume": 3,
    },
    "E": {
        "aliphaticity": 1,
        "aromaticity": 1,
        "composition": 1,
        "hydrogenation": 1,
        "hydropathy": 1,
        "hydroxythiolation": 1,
        "iep": 1,
        "polar requirement": 4,
        "polarity": 4,
        "volume": 4,
    },
    "F": {
        "aliphaticity": 1,
        "aromaticity": 2,
        "composition": 1,
        "hydrogenation": 1,
        "hydropathy": 3,
        "hydroxythiolation": 3,
        "iep": 2,
        "polar requirement": 1,
        "polarity": 1,
        "volume": 4,
    },
    "G": {
        "aliphaticity": 1,
        "aromaticity": 1,
        "composition": 1,
        "hydrogenation": 1,
        "hydropathy": 2,
        "hydroxythiolation": 2,
        "iep": 2,
        "polar requirement": 2,
        "polarity": 2,
        "volume": 1,
    },
    "H": {
        "aliphaticity": 1,
        "aromaticity": 2,
        "composition": 1,
        "hydrogenation": 1,
        "hydropathy": 1,
        "hydroxythiolation": 3,
        "iep": 3,
        "polar requirement": 2,
        "polarity": 3,
        "volume": 4,
    },
    "I": {
        "aliphaticity": 1,
        "aromaticity": 1,
        "composition": 1,
        "hydrogenation": 1,
        "hydropathy": 4,
        "hydroxythiolation": 1,
        "iep": 2,
        "polar requirement": 1,
        "polarity": 1,
        "volume": 4,
    },
    "K": {
        "aliphaticity": 1,
        "aromaticity": 1,
        "composition": 1,
        "hydrogenation": 1,
        "hydropathy": 1,
        "hydroxythiolation": 1,
        "iep": 3,
        "polar requirement": 3,
        "polarity": 4,
        "volume": 4,
    },
    "L": {
        "aliphaticity": 1,
        "aromaticity": 1,
        "composition": 1,
        "hydrogenation": 1,
        "hydropathy": 4,
        "hydroxythiolation": 1,
        "iep": 2,
        "polar requirement": 1,
        "polarity": 1,
        "volume": 4,
    },
    "M": {
        "aliphaticity": 1,
        "aromaticity": 1,
        "composition": 1,
        "hydrogenation": 1,
        "hydropathy": 3,
        "hydroxythiolation": 3,
        "iep": 2,
        "polar requirement": 1,
        "polarity": 1,
        "volume": 4,
    },
    "N": {
        "aliphaticity": 1,
        "aromaticity": 1,
        "composition": 2,
        "hydrogenation": 1,
        "hydropathy": 1,
        "hydroxythiolation": 4,
        "iep": 2,
        "polar requirement": 3,
        "polarity": 4,
        "volume": 3,
    },
    "P": {
        "aliphaticity": 1,
        "aromaticity": 1,
        "composition": 1,
        "hydrogenation": 1,
        "hydropathy": 2,
        "hydroxythiolation": 2,
        "iep": 2,
        "polar requirement": 2,
        "polarity": 2,
        "volume": 2,
    },
    "Q": {
        "aliphaticity": 1,
        "aromaticity": 1,
        "composition": 1,
        "hydrogenation": 1,
        "hydropathy": 1,
        "hydroxythiolation": 2,
        "iep": 2,
        "polar requirement": 2,
        "polarity": 3,
        "volume": 4,
    },
    "R": {
        "aliphaticity": 1,
        "aromaticity": 1,
        "composition": 1,
        "hydrogenation": 1,
        "hydropathy": 1,
        "hydroxythiolation": 1,
        "iep": 3,
        "polar requirement": 2,
        "polarity": 3,
        "volume": 4,
    },
    "S": {
        "aliphaticity": 1,
        "aromaticity": 1,
        "composition": 2,
        "hydrogenation": 1,
        "hydropathy": 2,
        "hydroxythiolation": 5,
        "iep": 2,
        "polar requirement": 2,
        "polarity": 2,
        "volume": 2,
    },
    "T": {
        "aliphaticity": 1,
        "aromaticity": 1,
        "composition": 1,
        "hydrogenation": 1,
        "hydropathy": 2,
        "hydroxythiolation": 5,
        "iep": 2,
        "polar requirement": 2,
        "polarity": 2,
        "volume": 3,
    },
    "V": {
        "aliphaticity": 1,
        "aromaticity": 1,
        "composition": 1,
        "hydrogenation": 1,
        "hydropathy": 4,
        "hydroxythiolation": 1,
        "iep": 2,
        "polar requirement": 1,
        "polarity": 1,
        "volume": 4,
    },
    "W": {
        "aliphaticity": 1,
        "aromaticity": 2,
        "composition": 1,
        "hydrogenation": 1,
        "hydropathy": 2,
        "hydroxythiolation": 3,
        "iep": 2,
        "polar requirement": 1,
        "polarity": 1,
        "volume": 5,
    },
    "Y": {
        "aliphaticity": 1,
        "aromaticity": 2,
        "composition": 1,
        "hydrogenation": 1,
        "hydropathy": 2,
        "hydroxythiolation": 4,
        "iep": 2,
        "polar requirement": 1,
        "polarity": 1,
        "volume": 4,
    },
}



from unittest import TestCase
from dark.analyze_reads import getPrefixAndSuffix, trimReads
from Bio import SeqIO
from io import StringIO


class TestGetPrefixAndSuffix(TestCase):
    """
    Tests for getPrefixAndSuffix()
    """

    def testZeroInput(self):
        result = getPrefixAndSuffix(StringIO())
        self.assertEqual(result, (0, 0))

    def testOneInput(self):
        seq = ">hey\nagtcagtcagtc"
        result = getPrefixAndSuffix(StringIO(seq))
        self.assertEqual(result, (12, 12))

    def testTwoDifferentReads(self):
        seq = ">hey\nagtcagtcagtc\n>you\ntcctg"
        result = getPrefixAndSuffix(StringIO(seq))
        self.assertEqual(result, (0, 0))

    def testSamePrefixTwoReads(self):
        seq = ">hey\nagtcagtcagtc\n>you\nagttcctg"
        result = getPrefixAndSuffix(StringIO(seq))
        self.assertEqual(result, (3, 0))

    def testSameSuffixTwoReads(self):
        seq = ">hey\nagtcagtcagtc\n>you\ntcctggtc"
        result = getPrefixAndSuffix(StringIO(seq))
        self.assertEqual(result, (0, 3))

    def testSamePrefixAndSuffixTwoReads(self):
        seq = ">hey\nagtcagtcagtc\n>you\nagttcctggtc"
        result = getPrefixAndSuffix(StringIO(seq))
        self.assertEqual(result, (3, 3))

    def testSamePrefixDifferentSuffixThreeReads(self):
        seq = ">hey\nagtcagtcagtc\n>you\nagttcctggtc\n>how\nagtcggtat"
        result = getPrefixAndSuffix(StringIO(seq))
        self.assertEqual(result, (3, 0))

    def testSamePrefixSameSuffixThreeReads(self):
        seq = ">hey\nagtccttagatcg\n>you\nagtcgaatcg\n>how\nagtaacctcg"
        result = getPrefixAndSuffix(StringIO(seq))
        self.assertEqual(result, (3, 3))

    def testDifferentPrefixSameSuffixThreeReads(self):
        seq = ">hey\nagtccttagatcg\n>you\ncgaatcg\n>how\natgacctcg"
        result = getPrefixAndSuffix(StringIO(seq))
        self.assertEqual(result, (0, 3))


class TestTrimReads(TestCase):
    """
    Tests for trimReads()
    """

    def testZeroInput(self):
        result = trimReads(0, 0, StringIO())
        test_result = list(SeqIO.parse(StringIO(), "fasta"))
        self.assertEqual(
            list(map(str, tuple(result))), list(map(str, tuple(test_result)))
        )

    def testOneInput(self):
        result = trimReads(10, 10, StringIO())
        test_result = list(SeqIO.parse(StringIO(), "fasta"))
        self.assertEqual(
            list(map(str, tuple(result))), list(map(str, tuple(test_result)))
        )

    def testZeroPrefixZeroSuffix(self):
        seq = ">hey\nagtccgatcg"
        result = trimReads(0, 0, StringIO(seq))
        trimmed_seq = ">hey\nagtccgatcg"
        test_result = list(SeqIO.parse(StringIO(trimmed_seq), "fasta"))
        self.assertEqual(
            list(map(str, tuple(result))), list(map(str, tuple(test_result)))
        )

    def testPrefixZeroSuffix(self):
        seq = ">hey\nagtccgatcg"
        trimmed_seq = ">hey\nccgatcg"
        result = trimReads(3, 0, StringIO(seq))
        test_result = list(SeqIO.parse(StringIO(trimmed_seq), "fasta"))
        self.assertEqual(
            list(map(str, tuple(result))), list(map(str, tuple(test_result)))
        )

    def testZeroPrefixSuffix(self):
        seq = ">hey\nagtccgatcg"
        trimmed_seq = ">hey\nagtccga"
        result = trimReads(0, 3, StringIO(seq))
        test_result = list(SeqIO.parse(StringIO(trimmed_seq), "fasta"))
        self.assertEqual(
            list(map(str, tuple(result))), list(map(str, tuple(test_result)))
        )

    def testPrefixSuffix(self):
        seq = ">hey\nagtccgatcg"
        trimmed_seq = ">hey\nccga"
        result = trimReads(3, 3, StringIO(seq))
        test_result = list(SeqIO.parse(StringIO(trimmed_seq), "fasta"))
        self.assertEqual(
            list(map(str, tuple(result))), list(map(str, tuple(test_result)))
        )

    def testLongPrefixSuffix(self):
        seq = ">hey\nagtccgatcg"
        trimmed_seq = ">hey\n"
        result = trimReads(8, 4, StringIO(seq))
        test_result = list(SeqIO.parse(StringIO(trimmed_seq), "fasta"))
        self.assertEqual(
            list(map(str, tuple(result))), list(map(str, tuple(test_result)))
        )

    def testPrefixLongSuffix(self):
        seq = ">hey\nagtccgatcg"
        trimmed_seq = ">hey\n"
        result = trimReads(4, 8, StringIO(seq))
        test_result = list(SeqIO.parse(StringIO(trimmed_seq), "fasta"))
        self.assertEqual(
            list(map(str, tuple(result))), list(map(str, tuple(test_result)))
        )



from os import unlink
from tempfile import TemporaryFile
from resource import getrlimit, RLIMIT_NOFILE

from dark.btop import btop2cigar
from dark.diamond.conversion import DiamondTabularFormat
from dark.fpcache import FilePointerCache
from dark.reads import DNARead
from dark.utils import asHandle


class SAMWriter:
    """
    A class that collects SAM output lines and writes SAM to a file.

    @param ram: If C{True}, do not use a temporary file to hold the non-header
        SAM output. This will run faster but use more memory since all
        non-header SAM output will be stored in RAM and only written out when
        the full header can be determined.
    """

    def __init__(self, ram=False):
        self._ram = ram
        self._referenceLengths = {}

    def addMatchLine(self, line, genome, fp):
        """
        Add a line of SAM output.

        @param line: A C{str} line of valid pre-formatted SAM output.
        @param genome: A C{dict} with information about the (nucleotide) genome
            that the protein in the DIAMOND match comes from. The C{dict} is
            as returned by C{dark.proteins.SqliteIndex.findGenome}.
        @param fp: The file pointer to write the line to.
        """
        genomeName = genome["accession"]
        genomeLength = genome["length"]
        try:
            preexistingLength = self._referenceLengths[genomeName]
        except KeyError:
            self._referenceLengths[genomeName] = genomeLength
        else:
            if preexistingLength != genomeLength:
                raise ValueError(
                    "Reference %r passed with length %d but has already been "
                    "given with a different length (%d)."
                    % (genomeName, genomeLength, preexistingLength)
                )

        print(line, file=fp)

    def save(self, fp, tmpfp=None, headerLines=None):
        """
        Write SAM output to a file.

        @param fp: An open file pointer to write the full SAM to.
        @param tmpfp: An open file pointer with the non-header SAM output.
            Only used if C{self._ram} is C{False}
        @param headerLines: An optional iterable of C{str}s SAM header
            lines (with no trailing newlines).
        """
        # Print SAM headers.
        print(
            "\n".join(
                (headerLines or [])
                + [
                    "@SQ\tSN:%s\tLN:%d" % (name, self._referenceLengths[name])
                    for name in sorted(self._referenceLengths)
                ]
            ),
            file=fp,
        )

        # Print SAM match lines.
        if self._ram:
            # Trade speed for memory and print the non-header lines one at a
            # time instead of making and printing a potentially massive string
            # with '\n'.join() (although arguably we could just make the single
            # string, seeing as self._ram is True.)
            for line in self._nonHeaderLines:
                print(line, file=fp)
        else:
            tmpfp.seek(0)
            for line in tmpfp:
                print(line, end="", file=fp)


class _DiamondSAMWriter:
    """
    A base class for converting DIAMOND tabular output to SAM.

    @param genomesProteins: An open sqlite3 database with information on
        genomes and proteins, as made by the dark-matter
        make-protein-database.py script.
    @param mappingQuality: The mapping quality to use for the SAM MAPQ field
        (#5). The default (255) indicates that mapping quality information is
        not available.
    @param ram: If C{True}, do not use a temporary file to hold the non-header
        SAM output. This will run faster but use more memory since all
        non-header SAM output will be stored in RAM and only written out when
        the full header can be determined.
    @param keepDescriptions: If C{True}, do not discard text after the first
        space in query or subject sequence ids. Note that this violates the
        SAM specification, but since SAM files are TAB-separated there may be
        only a small chance this will cause problems downstream.
    """

    def __init__(
        self, genomesProteins, mappingQuality=255, ram=False, keepDescriptions=False
    ):
        self._genomesProteins = genomesProteins
        self._mappingQuality = mappingQuality
        self._ram = ram
        self._keepDescriptions = keepDescriptions
        self._strToDiamondDict = DiamondTabularFormat().diamondFieldsToDict
        self._referenceLengths = {}

    def _preprocessMatch(self, diamondStr):
        """
        Add information from a row of DIAMOND tabular output.

        @param diamondStr: A C{str} with TAB-separated fields from DIAMOND
            output format 6.
        @return: A C{tuple} containing the match C{dict} (from
            C{self._strToDiamondDict}), the protein C{dict} as looked up
            by C{self._genomesProteins.findProtein}, and the genome C{dict}
            as looked up by C{self._genomesProteins.findGenome}.
        """
        match = self._strToDiamondDict(diamondStr)
        stitle = match["stitle"]

        genomeAccession = self._genomesProteins.genomeAccession(stitle)
        genome = self._genomesProteins.findGenome(genomeAccession)
        if genome is None:
            raise ValueError(
                "Could not find accession %r in genomes database table."
                % genomeAccession
            )

        proteinAccession = self._genomesProteins.proteinAccession(stitle)
        protein = self._genomesProteins.findProtein(proteinAccession)
        if protein is None:
            raise ValueError(
                "Could not find accession %r in proteins database table."
                % proteinAccession
            )

        self._referenceLengths[genomeAccession] = genome["length"]

        return match, protein, genome

    def _SAMLine(self, match, protein, genome):
        """
        Convert DIAMOND match information to a line of SAM file output.

        @param match: A C{dict} with information about the DIAMOND match, as
            returned by C{DiamondTabularFormat().diamondFieldsToDict} which
            has been called for us by C{self._preprocessMatch}.
        @param protein: A C{dict} with information about the protein the
            DIAMOND was for. The C{dict} is as returned by
            C{dark.proteins.SqliteIndex.findProtein}.
        @param genome: A C{dict} with information about the (nucleotide) genome
            that the protein in the DIAMOND match comes from. The C{dict} is
            as returned by C{dark.proteins.SqliteIndex.findGenome}.
        @return: A TAB-separated C{str} line of SAM.
        """
        qseqid = (
            match["qseqid"]
            if self._keepDescriptions
            else match["qseqid"].split(None, 1)[0]
        )

        # from dark.genbank import GenomeRanges
        # ranges = GenomeRanges(protein['offsets'])

        # matchStartInGenome = ranges.startInGenome(match)
        # queryStartInGenome = matchStartInGenome - match['qstart'] - 1
        # orientations = ranges.orientations()

        # print('ranges', ranges)
        # print('protein', protein)
        # print('match', match)

        # If the query frame is less than zero, the match was with a
        # translation of the reverse-complemented query. We'll put the
        # reverse complement into the SAM output. This seems to be standard
        # / accepted practice, based on my web searches.  See e.g.,
        # https://www.biostars.org/p/131891/ for what Bowtie2 does and for
        # some comments on this issue for SAM/BAM files in general.
        if match["qframe"] > 0:
            flag = 0
            qseq = match["full_qseq"]
            qqual = match["full_qqual"] or "*"
        else:
            flag = 16
            qseq = DNARead("id", match["full_qseq"]).reverseComplement().sequence
            qqual = match["full_qqual"][::-1] if match["full_qqual"] else "*"

        # Make a CIGAR string, including hard-clipped bases at the start and
        # end of the query (DIAMOND outputs a hard-clipped query sequence).
        startClipCount = match["qstart"] - 1
        endClipCount = match["qlen"] - match["qend"]

        assert startClipCount >= 0
        assert (
            endClipCount >= 0
        ), "Query sequence %s has length %d but the qend value is %d" % (
            qseq,
            len(match["qseq"]),
            match["qend"],
        )

        cigar = (
            ("%dH" % startClipCount if startClipCount else "")
            + "".join(btop2cigar(match["btop"], concise=False, aa=True))
            + ("%dH" % endClipCount if endClipCount else "")
        )

        return "\t".join(
            map(
                str,
                (
                    # 1. QNAME
                    qseqid,
                    # 2. FLAG
                    flag,
                    # 3. RNAME
                    genome["accession"],
                    # 4. POS. This needs to be a 1-based offset into the
                    # nucleotide-equivalent of the DIAMOND subject sequence (which was
                    # a protein since that is how DIAMOND operates). Because DIAMOND
                    # gives back a 1-based protein location, we adjust to 0-based,
                    # multiply by 3 to get to nucleotides, then adjust to 1-based.
                    3 * (match["sstart"] - 1) + 1,
                    # 5. MAPQ
                    self._mappingQuality,
                    # 6. CIGAR
                    cigar,
                    # 7. RNEXT
                    "*",
                    # 8. PNEXT
                    0,
                    # 9. TLEN
                    0,
                    # 10. SEQ
                    qseq,
                    # 11. QUAL
                    qqual,
                    # 12. Alignment score
                    "AS:i:%d" % int(match["bitscore"]),
                ),
            )
        )

    def addMatch(self, diamondStr):
        """
        Add information from a row of DIAMOND tabular output.

        @param diamondStr: A C{str} with TAB-separated fields from DIAMOND
            output format 6.
        """
        raise NotImplementedError("addMatch must be implemented by a subclass")

    def save(self, fp):
        """
        Write SAM output to a file.
        """
        raise NotImplementedError("save must be implemented by a subclass")


class SimpleDiamondSAMWriter(_DiamondSAMWriter):
    """
    Convert DIAMOND tabular output to SAM.

    @param genomesProteins: An open sqlite3 database with information on
        genomes and proteins, as made by the dark-matter
        make-protein-database.py script.
    @param mappingQuality: The mapping quality to use for the SAM MAPQ field
        (#5). The default (255) indicates that mapping quality information is
        not available.
    @param ram: If C{True}, do not use a temporary file to hold the non-header
        SAM output. This will run faster but use more memory since all
        non-header SAM output will be stored in RAM and only written out when
        the full header can be determined.
    @param keepDescriptions: If C{True}, do not discard text after the first
        space in query or subject sequence ids. Note that this violates the
        SAM specification, but since SAM files are TAB-separated there may be
        only a small chance this will cause problems downstream.
    """

    def __init__(
        self, genomesProteins, mappingQuality=255, ram=False, keepDescriptions=False
    ):
        _DiamondSAMWriter.__init__(
            self,
            genomesProteins,
            mappingQuality=mappingQuality,
            ram=ram,
            keepDescriptions=keepDescriptions,
        )
        self._tf = TemporaryFile(mode="w+t", encoding="utf-8")
        self._writer = SAMWriter(ram=ram)

    def addMatch(self, diamondStr):
        """
        Add information from a row of DIAMOND tabular output.

        @param diamondStr: A C{str} with TAB-separated fields from DIAMOND
            output format 6.
        """
        match, protein, genome = self._preprocessMatch(diamondStr)
        self._writer.addMatchLine(
            self._SAMLine(match, protein, genome), genome, self._tf
        )

    def save(self, filename):
        """
        Write SAM output to a file.

        @param filename: A C{str} filename or an already-open file handle.
        """
        with asHandle(filename) as fp:
            self._writer.save(fp, tmpfp=self._tf)
        self._tf.close()
        self._tf = self._writer = None


class PerReferenceDiamondSAMWriter(_DiamondSAMWriter):
    """
    Convert DIAMOND tabular output to SAM and write it to per-reference
    output files.

    @param genomesProteins: An open sqlite3 database with information on
        genomes and proteins, as made by the dark-matter
        make-protein-database.py script.
    @param baseFilenameFunc: A function that produces a base filename to save
        SAM output to for a given reference name. A '.sam' suffix will be
        added to these files.
    @param mappingQuality: The mapping quality to use for the SAM MAPQ field
        (#5). The default (255) indicates that mapping quality information is
        not available.
    @param ram: If C{True}, do not use a temporary file to hold the non-header
        SAM output. This will run faster but use more memory since all
        non-header SAM output will be stored in RAM and only written out when
        the full header can be determined.
    @param keepDescriptions: If C{True}, do not discard text after the first
        space in query or subject sequence ids. Note that this violates the
        SAM specification, but since SAM files are TAB-separated there may be
        only a small chance this will cause problems downstream.
    @param fpcMaxsize: An C{int} maximum size for the file pointer cache used
        because SAM output may be being written to an arbitrary number of
        output files so we need to make sure the OS limit on open files is not
        exceeded and that we are not inefficiently constantly re-opening and
        writing single lines to files). If C{None} or C{0} is passed, a size
        of half the maximum number of files will be used.
    """

    SAM_SUFFIX = ".sam"
    TMP_SUFFIX = ".tmp"

    def __init__(
        self,
        genomesProteins,
        baseFilenameFunc=None,
        mappingQuality=255,
        ram=False,
        keepDescriptions=False,
        fpcMaxsize=0,
    ):
        _DiamondSAMWriter.__init__(
            self,
            genomesProteins,
            mappingQuality=mappingQuality,
            ram=ram,
            keepDescriptions=keepDescriptions,
        )

        self._writers = {}
        self._baseFilenameFunc = baseFilenameFunc
        self._fpc = FilePointerCache(
            maxsize=(fpcMaxsize or getrlimit(RLIMIT_NOFILE)[0] >> 1),
            openArgs={"mode": "wt"},
            reopenArgs={"mode": "at"},
        )

    def addMatch(self, diamondStr):
        """
        Add information from a row of DIAMOND tabular output.

        @param diamondStr: A C{str} with TAB-separated fields from DIAMOND
            output format 6.
        """
        match, protein, genome = self._preprocessMatch(diamondStr)
        genomeAccession = genome["accession"]
        basename = self._baseFilenameFunc(match["stitle"])

        try:
            writer, _ = self._writers[genomeAccession]
        except KeyError:
            writer = SAMWriter()
            self._writers[genomeAccession] = (writer, basename)

        writer.addMatchLine(
            self._SAMLine(match, protein, genome),
            genome,
            self._fpc.open(basename + self.TMP_SUFFIX),
        )

    def save(self):
        """
        Write SAM output to all files.
        """
        # Close the open file descriptor cache.
        self._fpc.close()

        # Write out SAM (including headers) for each reference.
        for writer, basename in self._writers.values():
            tmp = basename + self.TMP_SUFFIX
            with open(basename + self.SAM_SUFFIX, "wt") as samfp:
                with open(tmp) as tmpfp:
                    writer.save(samfp, tmpfp)
            unlink(tmp)



#!/usr/bin/env python

from Bio import SeqIO
from collections import defaultdict
import sys
from dark.distance import levenshtein
from math import log10, ceil

# The name of the unknown adaptor.
UNKNOWN = "UNKNOWN"


def splitFASTAByAdaptor(
    knownAdaptors,
    adaptorLen,
    adaptorOffset,
    maximumDistance,
    outputPrefix,
    dryRun,
    verbose,
):
    """
    @param knownAdaptors: A C{set} of expected adaptor sequences.
    @param adaptorLen: The C{int} length of each adaptor sequence.
    @param adaptorOffset: The zero-based C{int} offset of the adaptor in
        each sequence.
    @param maximumDistance: The maximum distance an unknown adaptor will be
        mapped to in an attempt to find its nearest known adaptor.
    @param outputPrefix: A C{str} prefix that should be used in the file names
        that are written out.
    @param dryRun: A C{bool}, if C{True} only print what would be done, don't
        create any new FASTA files.
    @param verbose: A C{bool}, if C{True} output additional information about
        adaptor classes found and assigned.
    """
    adaptors = defaultdict(int)
    unknowns = 0
    classes = dict(zip(knownAdaptors, knownAdaptors))
    reads = []

    for count, seq in enumerate(SeqIO.parse(sys.stdin, "fasta"), start=1):
        reads.append(seq)
        adaptor = str(seq.seq)[adaptorOffset:][:adaptorLen].upper()
        adaptors[adaptor] += 1

    order = sorted(adaptors, key=lambda adaptor: adaptors[adaptor], reverse=True)

    for adaptor in order:
        if adaptor in knownAdaptors:
            if verbose:
                print("%s: %s. Known adaptor" % (adaptor, adaptors[adaptor]))
        else:
            distances = sorted(
                (levenshtein(adaptor, known), known) for known in knownAdaptors
            )
            # Treat the read as unclassifiable if it's too far from its
            # nearest neighbor or if its nearest neighbor is ambiguous.
            nearest = distances[0][0]
            if nearest > maximumDistance or (
                len(knownAdaptors) > 1 and nearest == distances[1][0]
            ):
                unknowns += 1
                classes[adaptor] = UNKNOWN
                if verbose:
                    print(
                        "%s: %s. Unknown, distances %r"
                        % (adaptor, adaptors[adaptor], [d[0] for d in distances])
                    )
            else:
                correctedAdaptor = distances[0][1]
                classes[adaptor] = correctedAdaptor
                if verbose:
                    print(
                        "%s: %s. Assigned to class %s, at dist %d"
                        % (
                            adaptor,
                            adaptors[adaptor],
                            correctedAdaptor,
                            distances[0][0],
                        )
                    )

    readGroups = defaultdict(list)

    # Collect reads into classes.
    for read in reads:
        adaptor = str(read.seq)[adaptorOffset:][:adaptorLen].upper()
        readGroups[classes[adaptor]].append(read[adaptorOffset + adaptorLen :])

    # Calculate the number of digits in the size of the biggest read group
    # so we can nicely align the output.
    width = int(ceil(log10(max(len(group) for group in readGroups.values()))))

    # The width of the count of files we'll write, so file names have zero
    # padded numeric prefixes.
    filesWidth = int(ceil(log10(len(readGroups))))

    # Write out the FASTA files for each adaptor class (this includes the
    # unclassifiable reads if any unknown adaptors were found).
    for count, adaptor in enumerate(sorted(readGroups), start=1):
        reads = readGroups[adaptor]
        filename = "%s%0*d-%s.fasta" % (outputPrefix, filesWidth, count, adaptor)
        description = (
            "unrecognized adaptors" if adaptor == UNKNOWN else "adaptor %s" % adaptor
        )
        if dryRun:
            print(
                "Would write %*d sequences for %s to %s"
                % (width, len(reads), description, filename)
            )
        else:
            with open(filename, "w") as fp:
                SeqIO.write(reads, fp, "fasta")
            print(
                "Wrote %*d sequences for %s to %s"
                % (width, len(reads), description, filename)
            )


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description=(
            "From a set of known adaptors and a FASTA on stdin, "
            "write FASTA files split by adaptor"
        )
    )

    parser.add_argument(
        "adaptors",
        nargs="+",
        metavar="adaptor",
        help="the set of adaptors that were used in sequencing",
    )

    parser.add_argument(
        "--adaptor-offset",
        metavar="N",
        type=int,
        default=0,
        dest="adaptorOffset",
        help="the offset of the adaptor in each sequence",
    )

    parser.add_argument(
        "--maximum-distance",
        metavar="dist",
        type=int,
        default=2,
        dest="maximumDistance",
        help="The maximum distance an unknown adaptor will be "
        "mapped to in an attempt to find its nearest known adaptor.",
    )

    parser.add_argument(
        "--outputPrefix",
        metavar="prefix",
        default="MID-",
        help="the prefix to use for newly created FASTA files",
    )

    parser.add_argument(
        "--dry-run",
        type=bool,
        default=False,
        dest="dryRun",
        help="If True, do not write new FASTA files, just show what would be " "done",
    )

    parser.add_argument(
        "--verbose",
        type=bool,
        default=False,
        help="If True, print information about adaptor classes",
    )

    args = parser.parse_args()
    adaptorLen = len(args.adaptors[0])
    adaptors = set(args.adaptors)

    # Check all adaptors are the same length.
    if any((len(adaptor) != adaptorLen) for adaptor in adaptors):
        print("All adaptors must be the same length.", file=sys.stderr)
        sys.exit(1)

    splitFASTAByAdaptor(
        adaptors,
        adaptorLen,
        args.adaptorOffset,
        args.maximumDistance,
        args.outputPrefix,
        args.dryRun,
        args.verbose,
    )



#!/usr/bin/env python

import argparse
import bz2file
import sys

from dark.blast.conversion import XMLRecordsReader


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Convert a BLAST XML file to JSON.",
        epilog=(
            "Give a BLAST XML file and convert it to JSON. Optionally "
            "compress the JSON output."
        ),
    )

    parser.add_argument(
        "--json",
        metavar="JSON-output-file",
        help=(
            "the JSON filename to write the converted XML to. If omitted, "
            "standard output will be used."
        ),
    )

    parser.add_argument(
        "--xml",
        metavar="BLAST-XML-file",
        default=sys.stdin,
        help=(
            "the BLAST XML output file to convert. If omitted, standard "
            "input will be read."
        ),
    )

    parser.add_argument(
        "--bzip2",
        default=False,
        action="store_true",
        help="If True, compress the JSON output using bzip2.",
    )

    args = parser.parse_args()

    if args.bzip2:
        fp = bz2file.BZ2File(args.json or sys.stdout, "w")
    else:
        fp = open(args.json, "w") if args.json else sys.stdout

    reader = XMLRecordsReader(args.xml)
    reader.saveAsJSON(fp)
    fp.close()



#!/usr/bin/env python

import sys
import argparse
from Bio import SeqIO


parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description="Parse GenBank flat files. Exit non-zero if parsing fails.",
)

parser.add_argument(
    "--expectedCount",
    type=int,
    help=(
        "The expected number of records in each GenBank file. If given "
        "and this number is not present in any file, exit non-zero."
    ),
)

parser.add_argument("files", nargs="+", help="The files to parse.")

parser.add_argument("--quiet", action="store_true", help="If given, write no output.")

args = parser.parse_args()

expectedCount = args.expectedCount
totalCount = 0

for i in args.files:
    count = 0
    try:
        records = SeqIO.parse(open(i), "gb")
        for record in records:
            count += 1
    except ValueError:
        if not args.quiet:
            print("Could not parse %s" % i, file=sys.stderr)
        sys.exit(1)
    else:
        if not args.quiet:
            print("Read %d records from %s" % (count, i), file=sys.stderr)
        if expectedCount is not None and count != expectedCount:
            if not args.quiet:
                print("Expected %d records. Exiting." % expectedCount, file=sys.stderr)
            sys.exit(1)
        totalCount += count

if not args.quiet:
    print("Total records read: %d" % totalCount, file=sys.stderr)



#!/usr/bin/env python

import sys
from re import compile
import argparse

from dark.blast.alignments import BlastReadsAlignments
from dark.titles import TitlesAlignments
from dark.fasta import FastaReads


def main(recordFilenames, fastaFilename, title, xRange, bitRange):
    """
    Print reads that match in a specified X-axis and bit score range.

    @param recordFilenames: A C{list} of C{str} file names contain results of a
        BLAST run, in JSON format.
    @param fastaFilename: The C{str} name of the FASTA file that was originally
        BLASTed.
    @param title: The C{str} title of the subject sequence, as output by BLAST.
    @param xRange: A (start, end) list of C{int}s, giving an X-axis range or
        C{None} if the entire X axis range should be printed.
    @param bitRange: A (start, end) list of C{int}s, giving a bit score range
        or C{None} if the entire bit score range should be printed.
    """
    reads = FastaReads(fastaFilename)
    blastReadsAlignments = BlastReadsAlignments(reads, recordFilenames)
    filtered = blastReadsAlignments.filter(
        whitelist=set([title]), negativeTitleRegex="."
    )
    titlesAlignments = TitlesAlignments(filtered)

    if title not in titlesAlignments:
        print("%s: Title %r not found in BLAST output" % (sys.argv[0], title))
        sys.exit(3)

    for titleAlignment in titlesAlignments[title]:
        for hsp in titleAlignment.hsps:
            if (
                xRange is None
                or (xRange[0] <= hsp.subjectEnd and xRange[1] >= hsp.subjectStart)
            ) and (bitRange is None or (bitRange[0] <= hsp.score.score <= bitRange[1])):
                print(
                    (
                        "query: %s, start: %d, end: %d, score: %d"
                        % (
                            titleAlignment.read.id,
                            hsp.subjectStart,
                            hsp.subjectEnd,
                            hsp.score.score,
                        )
                    )
                )


if __name__ == "__main__":
    if len(sys.argv) < 3:
        print(
            (
                "Usage: %s recordFilename, fastaFilename, "
                "title, xCoords, bitCoords" % sys.argv[0]
            ),
            file=sys.stderr,
        )
        sys.exit(1)

    parser = argparse.ArgumentParser(
        description=(
            "Print the reads that are " "at specified positions in an alignmentGraph"
        ),
        epilog=(
            "Given a JSON BLAST output file, a title and an x and / or "
            "bitRange, print the reads that are within the given Ranges."
        ),
    )

    parser.add_argument(
        "json",
        metavar="BLAST-JSON-file",
        nargs="+",
        help="the JSON file of BLAST output.",
    )

    parser.add_argument(
        "fasta", metavar="fastaFile", help="the FASTA file of BLAST input."
    )

    parser.add_argument(
        "title", metavar="SEQUENCE-TITLE", help="The title of the subject sequence."
    )

    parser.add_argument("--xRange", default=None, help="a range on the x-axis.")

    parser.add_argument(
        "--bitRange", default=None, help="a bit score range on the y-axis."
    )

    args = parser.parse_args()

    def _getRange(inputRange):
        """
        Convert a string input range into a pair of integers.

        @param inputRange: A C{str}, either  a single number like '10', or a
            hyphen-separated pair of numbers like '3-9'. May also be C{None}.
        @return: Either C{None} if inputRange is C{None} or the empty string,
            else a (start, end) list of C{int}s.
        """
        if inputRange:
            rangeRegex = compile(r"^(\d+)(?:-(\d+))?$")
            match = rangeRegex.match(inputRange)
            if match:
                start, end = match.groups()
                start = int(start)
                if end is None:
                    end = start
                else:
                    end = int(end)
                if start > end:
                    start, end = end, start
                return start, end
            else:
                print(
                    (
                        "Illegal argument %r. Ranges must single numbers or "
                        "number-number." % inputRange
                    ),
                    file=sys.stderr,
                )
                sys.exit(2)

    main(
        args.json,
        args.fasta,
        args.title,
        _getRange(args.xRange),
        _getRange(args.bitRange),
    )



#!/usr/bin/env python

import argparse
from collections import defaultdict, Counter

from dark.filter import (
    addFASTAFilteringCommandLineOptions,
    parseFASTAFilteringCommandLineOptions,
)
from dark.reads import Reads
from dark.sam import samfile, SAMFilter

parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description="Print SAM/BAM file coverage statistics.",
)

parser.add_argument(
    "--noFilter",
    default=False,
    action="store_true",
    help=(
        "Do not use our SAM filtering. Note that if you give this option, "
        "any filtering option (other than --referenceId) you also specify "
        "that is provided by the SAMFilter.addFilteringOptions will be "
        "silently ignored!"
    ),
)

addFASTAFilteringCommandLineOptions(parser)
SAMFilter.addFilteringOptions(parser, samfileIsPositional=True)

args = parser.parse_args()

if args.noFilter:
    # Do not do our custom SAM filtering.
    def filterRead(read):
        return True

else:

    def filterRead(read):
        return not (read.is_del or read.is_refskip) and samFilter.filterAlignment(
            read.alignment
        )


# We don't have a file of reads, we just want a read filter that we can use
# to filter the SAM file query sequences and to get reference lengths from.
reads = parseFASTAFilteringCommandLineOptions(args, Reads())
samFilter = SAMFilter.parseFilteringOptions(args, reads.filterRead)

coveredOffsets = defaultdict(Counter)
coveringReads = defaultdict(set)

with samfile(args.samfile) as sam:
    for column in sam.pileup():
        referenceOffset = column.reference_pos
        for read in column.pileups:
            if filterRead(read):
                referenceId = read.alignment.reference_name
                coveredOffsets[referenceId][referenceOffset] += 1
                coveringReads[referenceId].add(read.alignment.query_name)

referenceLengths = samFilter.referenceLengths()

for referenceId in sorted(referenceLengths):
    offsetsCovered = len(coveredOffsets[referenceId])
    referenceLength = referenceLengths[referenceId]
    print(
        "%s: length %d, covering reads %d, covered sites %d (%.4f%%), "
        "mean coverage depth %.4f (min: %d, max: %d)"
        % (
            referenceId,
            referenceLength,
            len(coveringReads[referenceId]),
            offsetsCovered,
            offsetsCovered / referenceLength * 100.0,
            sum(coveredOffsets[referenceId].values()) / referenceLength,
            min(coveredOffsets[referenceId].values(), default=0),
            max(coveredOffsets[referenceId].values(), default=0),
        )
    )
